# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'vector.ui'
#
# Created: Sat Oct 17 11:23:24 2015
# by: PyQt4 UI code generator 4.11.3
#
# WARNING! All changes made in this file will be lost!

import json
from PySide import QtCore, QtGui
from PySide.QtUiTools import QUiLoader
import pyqtgraph as pg
import pyqtgraph.parametertree.parameterTypes as pTypes
from pyqtgraph.parametertree import Parameter, ParameterTree, ParameterItem, registerParameterType
import numpy as np
from pyFAI import detectors
import hig
import featuremanager
import ui
import display
from modpkgs.collectionsmod import UnsortableOrderedDict

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    def _fromUtf8(s):
        return s

try:
    _encoding = QtGui.QApplication.UnicodeUTF8


    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig, _encoding)
except AttributeError:
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig)


class vector(QtGui.QWidget):
    sigValueChanged = QtCore.Signal()
    sigChanged = sigValueChanged

    def __init__(self):
        super(vector, self).__init__()

        self.makeValues()

        self.horizontalLayout = QtGui.QHBoxLayout(self)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName(_fromUtf8("horizontalLayout"))
        self.UnitCellVec1LeftParenthesis3D = QtGui.QLabel(self)
        font = QtGui.QFont()
        font.setPointSize(30)
        self.UnitCellVec1LeftParenthesis3D.setFont(font)
        self.UnitCellVec1LeftParenthesis3D.setObjectName(_fromUtf8("UnitCellVec1LeftParenthesis3D"))
        self.horizontalLayout.addWidget(self.UnitCellVec1LeftParenthesis3D)
        self.value1.setMinimum(-1000)
        self.value1.setMaximum(1000)
        self.value1.setSingleStep(1)
        self.value1.setProperty("value", 0.0)
        self.value1.setObjectName(_fromUtf8("value1"))
        self.horizontalLayout.addWidget(self.value1)
        self.UnitCellVec1Comma1 = QtGui.QLabel(self)
        self.UnitCellVec1Comma1.setObjectName(_fromUtf8("UnitCellVec1Comma1"))
        self.horizontalLayout.addWidget(self.UnitCellVec1Comma1)
        self.value2.setMinimum(-1000)
        self.value2.setMaximum(1000)
        self.value2.setSingleStep(1)
        self.value2.setProperty("value", 0.0)
        self.value2.setObjectName(_fromUtf8("value2"))
        self.horizontalLayout.addWidget(self.value2)
        self.UnitCellVec1Comma2 = QtGui.QLabel(self)
        self.UnitCellVec1Comma2.setObjectName(_fromUtf8("UnitCellVec1Comma2"))
        self.horizontalLayout.addWidget(self.UnitCellVec1Comma2)
        self.value3.setMinimum(-1000)
        self.value3.setMaximum(1000)
        self.value3.setSingleStep(1)
        self.value3.setProperty("value", 0.0)
        self.value3.setObjectName(_fromUtf8("value3"))
        self.horizontalLayout.addWidget(self.value3)
        self.UnitCellVec1RightParenthesis3D = QtGui.QLabel(self)
        self.UnitCellVec1RightParenthesis3D.setFont(font)
        self.UnitCellVec1RightParenthesis3D.setObjectName(_fromUtf8("UnitCellVec1RightParenthesis3D"))
        self.horizontalLayout.addWidget(self.UnitCellVec1RightParenthesis3D)

        self.UnitCellVec1LeftParenthesis3D.setText("(")
        self.UnitCellVec1Comma1.setText(",")
        self.UnitCellVec1Comma2.setText(",")
        self.UnitCellVec1RightParenthesis3D.setText(")")

        self.setValue([0,0,0])

        self.value1.valueChanged.connect(self.sigValueChanged)
        self.value2.valueChanged.connect(self.sigValueChanged)
        self.value3.valueChanged.connect(self.sigValueChanged)

    def makeValues(self):
        self.value1 = QtGui.QDoubleSpinBox()
        self.value2 = QtGui.QDoubleSpinBox()
        self.value3 = QtGui.QDoubleSpinBox()
        for v in [self.value1,self.value2,self.value3]:
            v.setDecimals(3)

    def value(self):
        return self.value1.value(), self.value2.value(), self.value3.value()

    def setValue(self, v):
        if v is None: return
        self.value1.setValue(v[0])
        self.value2.setValue(v[1])
        self.value3.setValue(v[2])

    def setEnabled(self, enabled):
        self.value1.setEnabled(enabled)
        self.value2.setEnabled(enabled)
        self.value3.setEnabled(enabled)

class intvector(vector):
    sigValueChanged = QtCore.Signal()
    sigChanged = sigValueChanged

    def makeValues(self):
        self.value1 = QtGui.QSpinBox()
        self.value2 = QtGui.QSpinBox()
        self.value3 = QtGui.QSpinBox()


class ROlineEdit(QtGui.QLineEdit):
    def __init__(self, *args, **kwargs):
        super(ROlineEdit, self).__init__(*args, **kwargs)
        self.setReadOnly(True)
        self.setFrame(False)

    def focusOutEvent(self, *args, **kwargs):
        super(ROlineEdit, self).focusOutEvent(*args, **kwargs)
        self.setReadOnly(True)
        self.setFrame(False)
        self.setCursor(QtCore.Qt.ArrowCursor)

    def mouseDoubleClickEvent(self, *args, **kwargs):
        super(ROlineEdit, self).mouseDoubleClickEvent(*args, **kwargs)
        self.setReadOnly(False)
        self.setFrame(True)
        self.setFocus()
        self.selectAll()


# Form implementation generated from reading ui file 'layer.ui'
#
# Created: Thu Oct 22 09:52:22 2015
# by: pyside-uic 0.2.15 running on PySide 1.2.2
#
# WARNING! All changes made in this file will be lost!


class featureWidget(QtGui.QWidget):
    sigUpdateDisplayUnitCell = QtCore.Signal()

    def __init__(self, name=''):
        self.name = name

        self._form = None

        super(featureWidget, self).__init__()

        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Preferred, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sizePolicy().hasHeightForWidth())
        self.setSizePolicy(sizePolicy)
        self.verticalLayout = QtGui.QVBoxLayout(self)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.frame = QtGui.QFrame(self)
        self.frame.setFrameShape(QtGui.QFrame.StyledPanel)
        self.frame.setFrameShadow(QtGui.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.horizontalLayout_2 = QtGui.QHBoxLayout(self.frame)
        self.horizontalLayout_2.setSpacing(0)
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.pushButton = QtGui.QPushButton(self.frame)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushButton.sizePolicy().hasHeightForWidth())
        self.pushButton.setSizePolicy(sizePolicy)
        self.pushButton.setStyleSheet("margin:0 0 0 0;")
        self.pushButton.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_47.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.pushButton.setIcon(icon)
        self.pushButton.setFlat(True)
        self.pushButton.setObjectName("pushButton")
        self.horizontalLayout_2.addWidget(self.pushButton)
        self.line = QtGui.QFrame(self.frame)
        self.line.setFrameShape(QtGui.QFrame.VLine)
        self.line.setFrameShadow(QtGui.QFrame.Sunken)
        self.line.setObjectName("line")
        self.horizontalLayout_2.addWidget(self.line)
        # self.pushButton_2 = QtGui.QPushButton(self.frame)
        # sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Fixed)
        # sizePolicy.setHorizontalStretch(0)
        # sizePolicy.setVerticalStretch(0)
        # sizePolicy.setHeightForWidth(self.pushButton_2.sizePolicy().hasHeightForWidth())
        # self.pushButton_2.setSizePolicy(sizePolicy)
        # self.pushButton_2.setStyleSheet("margin:0 0 0 0;")
        # self.pushButton_2.setFlat(True)
        # self.pushButton_2.setObjectName("pushButton_2")
        # self.horizontalLayout_2.addWidget(self.pushButton_2)
        # self.line_2 = QtGui.QFrame(self.frame)
        # self.line_2.setFrameShape(QtGui.QFrame.VLine)
        # self.line_2.setFrameShadow(QtGui.QFrame.Sunken)
        # self.line_2.setObjectName("line_2")
        # self.horizontalLayout_2.addWidget(self.line_2)
        self.txtName = ROlineEdit(self.frame)
        self.txtName.setObjectName("txtName")
        self.horizontalLayout_2.addWidget(self.txtName)
        self.txtName.setText(name)
        self.line_3 = QtGui.QFrame(self.frame)
        self.line_3.setFrameShape(QtGui.QFrame.VLine)
        self.line_3.setFrameShadow(QtGui.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.horizontalLayout_2.addWidget(self.line_3)
        self.pushButton_3 = QtGui.QPushButton(self.frame)
        self.pushButton_3.setStyleSheet("margin:0 0 0 0;")
        self.pushButton_3.setText("")
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap("xicam/gui/icons_46.gif"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.pushButton_3.setIcon(icon1)
        self.pushButton_3.setFlat(True)
        self.pushButton_3.setObjectName("pushButton_3")
        self.pushButton_3.clicked.connect(self.delete)
        self.horizontalLayout_2.addWidget(self.pushButton_3)
        self.verticalLayout.addWidget(self.frame)

        self.txtName.mousePressEvent = self.mousePressEvent

        # self.pushButton_2.setText("O")

        self.frame.setFrameShape(QtGui.QFrame.Box)
        self.frame.setCursor(QtCore.Qt.ArrowCursor)

        _ = self.form  # force cache form

    def delete(self):
        value = QtGui.QMessageBox.question(None, 'Delete this feature?',
                                           'Are you sure you want to delete this feature?',
                                           (QtGui.QMessageBox.Yes | QtGui.QMessageBox.Cancel))
        if value is QtGui.QMessageBox.Yes:
            featuremanager.features.remove(self)
            ui.showForm(ui.blankForm)
            self.deleteLater()

        display.redraw()

    def mousePressEvent(self, *args, **kwargs):
        self.showSelf()
        self.hideothers()
        self.setFocus()
        super(featureWidget, self).mousePressEvent(*args, **kwargs)

    def showSelf(self):
        ui.showForm(self.form)

    def hideothers(self):
        for item in featuremanager.features:
            if hasattr(item, 'frame_2') and item is not self:
                item.frame_2.hide()

    @property
    def form(self):
        if self._form is None:
            self._form = loadform(self._formpath)
            self.wireup()
        return self._form

    def wireup(self):
        if hasattr(self.form, 'txtName'):
            self.form.txtName.setText(self.name)
            self.form.txtName.textChanged.connect(self.setName)

    def setName(self, name):
        self.name = name
        featuremanager.update()

    def toDict(self):
        return {}


class substrate(featureWidget):
    def __init__(self, name='Substrate'):
        self._formpath = 'xicam/gui/guiSubstrate.ui'
        super(substrate, self).__init__(name)

    @property
    def form(self):
        if self._form is None:
            self.Material = pTypes.ListParameter(type='list', name='Material',
                                                 values=['Custom...', 'Si (10 keV)', 'Polystyrene (10 keV)',
                                                         'PMMA (10 keV)', 'Air', 'Vacuum'], value=0)
            self.delta = pTypes.SimpleParameter(type='float', name='delta', value=4.60093997E-06, step=.01)
            self.beta = pTypes.SimpleParameter(type='float', name='beta', value=3.91313968E-08, step=.01)
            params = [self.Material, self.delta, self.beta]

            self.parameter = Parameter.create(name='params', type='group', children=params)

            self.parameterTree = ParameterTree()
            self.parameterTree.setParameters(self.parameter, showTop=False)
            self._form = self.parameterTree
            self.wireup()
        return self._form

    def wireup(self):

        self.Material.sigValueChanged.connect(self.materialChanged)
        self.setConnected(True)

    def setConnected(self, connect):
        for item in [self.delta, self.beta]:
            if connect:
                item.sigValueChanged.connect(self.settoCustom)
            else:
                item.sigValueChanged.disconnect(self.settoCustom)

    def settoCustom(self):
        self.Material.setValue(0)

    def materialChanged(self, _, material):
        # print material
        self.setConnected(False)
        if material == 'PMMA (10 keV)':
            self.delta.setValue(0.5)
            self.beta.setValue(0.5)
        if material == 'Si (10 keV)':
            self.delta.setValue(4.60093997E-06)
            self.beta.setValue(3.91313968E-08)

        self.setConnected(True)

    def toDict(self):
        return UnsortableOrderedDict([('key', 'substr'),
                                      ('name', self.name),
                                      ('order', -1),
                                      ('material', self.Material.value()),
                                      ('refindex', UnsortableOrderedDict([('delta', self.delta.value()),
                                                                          ('beta', self.beta.value())]))])


class layer(featureWidget):
    def __init__(self, name=None):
        self._formpath = 'xicam/gui/guiLayer.ui'
        if name is None:
            name = 'Layer ' + str(featuremanager.layercount() + 1)
        super(layer, self).__init__(name)

    @property
    def form(self):
        if self._form is None:
            self.Thickness = pTypes.SimpleParameter(type='float', name='Thickness', value=.0000001, step=1e-9,
                                                    suffix='m', siPrefix=True)
            self.Material = pTypes.ListParameter(type='list', name='Material',
                                                 values=['Custom...', 'Si (10 keV)', 'Polystyrene (10 keV)',
                                                         'PMMA (10 keV)', 'Air', 'Vacuum'], value=0)
            self.delta = pTypes.SimpleParameter(type='float', name='delta', value=4.60093997E-06, step=.01)
            self.beta = pTypes.SimpleParameter(type='float', name='beta', value=3.91313968E-08, step=.01)
            self.TransVec = VectorParameter(name='Translation')
            params = [self.Thickness, self.Material, self.delta, self.beta, self.TransVec]

            self.parameter = Parameter.create(name='params', type='group', children=params)

            self.parameterTree = ParameterTree()
            self.parameterTree.setParameters(self.parameter, showTop=False)
            self._form = self.parameterTree
            self.wireup()
        return self._form

    def wireup(self):

        self.Material.sigValueChanged.connect(self.materialChanged)
        self.setConnected(True)

    def setConnected(self, connect):
        for item in [self.delta, self.beta]:
            if connect:
                item.sigValueChanged.connect(self.settoCustom)
            else:
                item.sigValueChanged.disconnect(self.settoCustom)

    def settoCustom(self):
        self.Material.setValue(0)

    def materialChanged(self, _, material):
        # print material
        self.setConnected(False)
        if material == 'PMMA (10 keV)':
            self.delta.setValue(0.5)
            self.beta.setValue(0.5)

        self.setConnected(True)

    def toDict(self):
        return UnsortableOrderedDict([('key', self.name),
                                      ('order', featuremanager.features.index(self)),
                                      ('material', self.Material.value()),
                                      ('thickness', self.Thickness.value()),
                                      ('transvec', list(self.TransVec.value())),
                                      ('refindex', UnsortableOrderedDict([('delta', self.delta.value()),
                                                                          ('beta', self.beta.value())]))])


class particle(featureWidget):
    def __init__(self, name=None):
        if name is None:
            name = 'Particle ' + str(featuremanager.particlecount() + 1)
        super(particle, self).__init__(name)

        self.frame_2 = QtGui.QFrame(self)
        self.frame_2.setFrameShape(QtGui.QFrame.StyledPanel)
        self.frame_2.setFrameShadow(QtGui.QFrame.Raised)
        self.frame_2.setObjectName("frame_2")
        self.gridLayout = QtGui.QGridLayout(self.frame_2)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setSpacing(0)
        self.gridLayout.setObjectName("gridLayout")
        self.line_4 = QtGui.QFrame(self.frame_2)
        self.line_4.setFrameShape(QtGui.QFrame.HLine)
        self.line_4.setFrameShadow(QtGui.QFrame.Sunken)
        self.line_4.setObjectName("line_4")
        self.gridLayout.addWidget(self.line_4, 1, 2, 1, 1)
        self.pushButton_4 = QtGui.QPushButton(self.frame_2)
        self.pushButton_4.setFlat(True)
        self.pushButton_4.setObjectName("pushButton_4")
        self.pushButton_4.clicked.connect(self.showStructure)
        self.gridLayout.addWidget(self.pushButton_4, 0, 2, 1, 1)
        self.pushButton_5 = QtGui.QPushButton(self.frame_2)
        self.pushButton_5.setFlat(True)
        self.pushButton_5.setObjectName("pushButton_5")
        self.pushButton_5.clicked.connect(self.showEnsemble)
        self.gridLayout.addWidget(self.pushButton_5, 2, 2, 1, 1)
        spacerItem = QtGui.QSpacerItem(20, 20, QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 0, 0, 1, 1)
        self.line_6 = QtGui.QFrame(self.frame_2)
        self.line_6.setFrameShape(QtGui.QFrame.VLine)
        self.line_6.setFrameShadow(QtGui.QFrame.Sunken)
        self.line_6.setObjectName("line_6")
        self.gridLayout.addWidget(self.line_6, 0, 1, 1, 1)
        self.line_7 = QtGui.QFrame(self.frame_2)
        self.line_7.setFrameShape(QtGui.QFrame.VLine)
        self.line_7.setFrameShadow(QtGui.QFrame.Sunken)
        self.line_7.setObjectName("line_7")
        self.gridLayout.addWidget(self.line_7, 2, 1, 1, 1)

        self.verticalLayout.addWidget(self.frame_2)
        self.pushButton_4.setText("Structure")
        self.pushButton_5.setText("Ensemble")
        self.frame_2.hide()

        self.frame_2.setFrameShape(QtGui.QFrame.Box)

        self.ensemble = ensemble(self)
        self.structure = structure(self)

    @property
    def form(self):
        if self._form is None:
            # PARTICLE STYLE
            self.Type = pTypes.ListParameter(name='Particle type', values=['Sphere', 'Box', 'Cylinder', 'Pyramid'],
                                             value=0)
            self.Type.sigValueChanged.connect(self.typeChanged)

            # SPHERE
            self.Radius = DistParameter(name='Radius', higKey='radius', value=10)

            # BOX
            self.Length = DistParameter(name='Length (x)', higKey='xsize', value=10)
            self.Width = DistParameter(name='Width (y)', higKey='ysize', value=10)
            self.Height = DistParameter(name='Height (z)', higKey='height', value=10)

            # CYLINDER
            # (redundant)

            # PYRAMID
            self.BaseAngle = DistParameter(name='Angle with base (deg)', higKey='baseangle', value=15)

            # ROTATIONS

            self.XRotation = pTypes.SimpleParameter(name='X Rotation', value=0, type='float')
            self.YRotation = pTypes.SimpleParameter(name='Y Rotation', value=0, type='float')
            self.ZRotation = pTypes.SimpleParameter(name='Z Rotation', value=0, type='float')

            # Refractive index
            self.delta = pTypes.SimpleParameter(type='float', name='delta', value=4.60093997E-06, step=.01)
            self.beta = pTypes.SimpleParameter(type='float', name='beta', value=3.91313968E-08, step=.01)
            self.RefractiveIndex = pTypes.GroupParameter(name='Refractive Index', children=[self.delta, self.beta])

            self.hideAll()
            self.Radius.show()

            params = [self.Type, self.Radius, self.Length, self.Width, self.Height, self.BaseAngle,
                      self.XRotation, self.YRotation, self.ZRotation, self.RefractiveIndex]
            self.parameter = Parameter.create(name='params', type='group', children=params)

            self.parameterTree = ParameterTree()
            self.parameterTree.setParameters(self.parameter, showTop=False)

            self.parameter.sigTreeStateChanged.connect(display.redraw)

            self._form = self.parameterTree
        return self._form

    def hideAll(self):
        self.Radius.hide()
        self.Length.hide()
        self.Width.hide()
        self.Height.hide()
        self.BaseAngle.hide()

    def typeChanged(self, _, choice):
        self.hideAll()
        for param in self.relevantParams():
            param.show()

    def wireup(self):
        super(particle, self).wireup()
        self.form
        self.setConnected(True)

        self.LatticeChoice.sigValueChanged.connect(self.changeUnitCellType)

    def showStructure(self):
        ui.showForm(self.structure.form)

    def showEnsemble(self):
        ui.showForm(self.ensemble.form)

    def mousePressEvent(self, *args, **kwargs):
        super(particle, self).mousePressEvent(*args, **kwargs)

        self.frame_2.show()

    def relevantParams(self):
        choice = self.Type.value()
        if choice == 'Sphere':
            return [self.Radius]
        elif choice == 'Box':
            return [self.Length, self.Width, self.Height]
        elif choice == 'Cylinder':
            return [self.Radius, self.Height]
        elif choice == 'Pyramid':
            return [self.Length, self.Width, self.Height, self.BaseAngle]

    def toDict(self):

        return UnsortableOrderedDict([('name', self.Type.value().lower()),
                                      ('key', self.name),
                                      ('params', [param.toDict() for param in self.relevantParams()]),
                                      ('xrot',self.XRotation.value()),
                                      ('yrot', self.YRotation.value()),
                                      ('zrot', self.ZRotation.value()),
                                      ('refindex', {'delta': self.delta.value(), 'beta': self.beta.value()})])


class form(QtGui.QWidget):
    def __init__(self, name):
        super(form, self).__init__()
        self._form = None
        self.wireup()
        self.form

    @property
    def form(self):
        if self._form is None:
            self._form = loadform(self._formpath)
        return self._form

    def wireup(self):
        pass


class ensemble(form):
    def __init__(self, parent, name='Ensemble'):
        self.parent = parent
        self._formpath = 'xicam/gui/guiEnsemble.ui'
        super(ensemble, self).__init__(name)


        # def wireup(self):
        # super(ensemble, self).wireup()
        # self.form.addGrainButton.clicked.connect(self.addGrain)

        # def addGrain(self):
        # self.addChild(grain())
        #     featuresmodel.layoutChanged.emit()

    @property
    def form(self):
        if self._form is None:
            # self.Axis = pTypes.ListParameter(name='Axis', values=['None', 'x', 'y', 'z'], value=0)
            self.addRotationAction = pTypes.ActionParameter(name='Add Rotation')
            self.addRotationAction.sigActivated.connect(self.addRotation)

            params = [self.addRotationAction]

            self.parameter = Parameter.create(name='params', type='group', children=params)

            self.parameterTree = ParameterTree()
            self.parameterTree.setParameters(self.parameter, showTop=False)
            self._form = self.parameterTree
        return self._form

    def addRotation(self):
        n = len(self.parameter.children())
        self.parameter.addChild(RotationParameter(name='Rotation ' + str(n)))

    def changeAxis(self, _, choice):
        # print choice
        if choice == 'None':
            self.Rotation.hide()
        else:
            self.Rotation.show()

    def toDict(self):
        # check that all the distributions are the same
        allsame=True
        if len(self.parameter.children())>1:
            disttype=self.parameter.children()[1].AngleDist.DistributionChoice.value()
            for rot in [rot for rot in self.parameter.children() if type(rot) is RotationParameter]:
                if not rot.AngleDist.DistributionChoice.value() == disttype:
                    allsame = False
        else:
            return dict()

        if not allsame:
            from xicam import dialogs
            dialogs.infodialog("HipGISAXS currently supports only 1 type of orientation distribution at a time. Please edit the ensemble.","Too many distributions")
            raise NotImplementedError

        stat = {'Uniform':'range','Random':'random','Bragg':'bragg','Gaussian':'gaussian'}

        rots = [('rot' + str(self.parameter.children().index(rot)), rot.toDict()) for rot in self.parameter.children()
                if type(rot) is RotationParameter]
        rots = [('stat',stat[disttype])]+rots
        return UnsortableOrderedDict(rots)


class structure(form):
    sigUpdateDisplayUnitCell = QtCore.Signal()

    def __init__(self, parent, name='Structure'):
        self.parent = parent
        self._formpath = 'xicam/gui/guiStructure.ui'
        self.vector1 = vector()
        self.vector2 = vector()
        self.vector3 = vector()
        super(structure, self).__init__(name)

        # Testing

        # {'name': 'Basic parameter data types', 'type': 'group', 'children': [
        # {'name': 'Integer', 'type': 'int', 'value': 10},
        #         {'name': 'Float', 'type': 'float', 'value': 10.5, 'step': 0.1},
        #         {'name': 'String', 'type': 'str', 'value': "hi"},
        #         {'name': 'List', 'type': 'list', 'values': [1,2,3], 'value': 2},
        #         {'name': 'Named List', 'type': 'list', 'values': {"one": 1, "two": "twosies", "three": [3,3,3]}, 'value': 2},
        #         {'name': 'Boolean', 'type': 'bool', 'value': True, 'tip': "This is a checkbox"},
        #         {'name': 'Color', 'type': 'color', 'value': "FF0", 'tip': "This is a color button"},
        #         {'name': 'Gradient', 'type': 'colormap'},
        #         {'name': 'Subgroup', 'type': 'group', 'children': [
        #             {'name': 'Sub-param 1', 'type': 'int', 'value': 10},
        #             {'name': 'Sub-param 2', 'type': 'float', 'value': 1.2e6},
        #             ]},
        #         {'name': 'Text Parameter', 'type': 'text', 'value': 'Some text...'},
        #         {'name': 'Action Parameter', 'type': 'action'},
        #         ]},
        #     {'name': 'Numerical Parameter Options', 'type': 'group', 'children': [
        #         {'name': 'Units + SI prefix', 'type': 'float', 'value': 1.2e-6, 'step': 1e-6, 'siPrefix': True, 'suffix': 'V'},
        #         {'name': 'Limits (min=7;max=15)', 'type': 'int', 'value': 11, 'limits': (7, 15), 'default': -6},
        #         {'name': 'DEC stepping', 'type': 'float', 'value': 1.2e6, 'dec': True, 'step': 1, 'siPrefix': True, 'suffix': 'Hz'},
        #
        #         ]},
        #     {'name': 'Save/Restore functionality', 'type': 'group', 'children': [
        #         {'name': 'Save State', 'type': 'action'},
        #         {'name': 'Restore State', 'type': 'action', 'children': [
        #             {'name': 'Add missing items', 'type': 'bool', 'value': True},
        #             {'name': 'Remove extra items', 'type': 'bool', 'value': True},
        #             ]},
        #         ]},
        #     {'name': 'Extra Parameter Options', 'type': 'group', 'children': [
        #         {'name': 'Read-only', 'type': 'float', 'value': 1.2e6, 'siPrefix': True, 'suffix': 'Hz', 'readonly': True},
        #         {'name': 'Renamable', 'type': 'float', 'value': 1.2e6, 'siPrefix': True, 'suffix': 'Hz', 'renamable': True},
        #         {'name': 'Removable', 'type': 'float', 'value': 1.2e6, 'siPrefix': True, 'suffix': 'Hz', 'removable': True},
        #         ]},

    @property
    def form(self):
        if self._form is None:
            self.dimensions = pTypes.ListParameter(name='Dimensions', type='list', value='3D',
                                                   values=['1D', '2D', '3D'])

            # 1-D/2D type
            self.non3DType = pTypes.ListParameter(name='Type', type='list', values=['Paracrystal', 'Percus Yevick'])
            self.non3DType.hide()
            self.domainsize = pTypes.SimpleParameter(name='Domain Size', type='float', value=100)
            self.domainsize.hide()
            self.non3Dlattice = pTypes.ListParameter(name='Planar Lattice', type='list', values=['Hexagonal','Cubic'])
            self.non3Dlattice.hide()

            # 1-D
            self.xmean = pTypes.SimpleParameter(name='Mean', type='float', value=100)
            self.xstd = pTypes.SimpleParameter(name='Std', type='float', value=1)
            self.xspacing = hideableGroupParameter(name='X Spacing', children=[self.xmean, self.xstd])
            self.xspacing.hide()

            # 2-D
            self.ymean = pTypes.SimpleParameter(name='Mean', type='float', value=100)
            self.ystd = pTypes.SimpleParameter(name='Std', type='float', value=1)
            self.yspacing = hideableGroupParameter(name='Y Spacing', children=[self.ymean, self.ystd])
            self.yspacing.hide()

            # 3-D
            self.LatticeA = VectorParameter(name='A', value=(30, 0, 0))
            self.LatticeB = VectorParameter(name='B', value=(0, 30, 0))
            self.LatticeC = VectorParameter(name='C', value=(0, 0, 30))
            self.LatticeChoice = pTypes.ListParameter(name='Type', type='list',
                                                      values=["Custom...", "Simple Cubic", "Body Centered Cubic",
                                                              'Face Centered Cubic', 'Hexagonal Planar'])
            self.Repetition = IntVectorParameter(name='Repetition', value=(2, 2, 2), type=int,step=1)
            self.Scaling = pTypes.SimpleParameter(name='Scaling', value=1, type='float')
            self.Lattice = hideableGroupParameter(name='Lattice', children=[self.LatticeChoice,
                                                                            self.LatticeA,
                                                                            self.LatticeB,
                                                                            self.LatticeC])

            self.Basis = ScalableGroup(name='Basis', children=[VectorParameter(name='Point 1')])
            self.Position = VectorParameter(name='Position')
            self.iratio = pTypes.SimpleParameter(name='I Ratio', value=1, type='float')

            params = [self.dimensions,
                      self.non3DType,
                      self.non3Dlattice,
                      self.xspacing,
                      self.yspacing,
                      self.domainsize,
                      self.Lattice,
                      self.Repetition,
                      self.Scaling,
                      self.Basis,
                      self.Position,
                      self.iratio
                      ]

            self.parameter = Parameter.create(name='params', type='group', children=params)

            self.parameterTree = ParameterTree()
            self.parameterTree.setParameters(self.parameter, showTop=False)
            self._form = self.parameterTree

            self. parameter.sigTreeStateChanged.connect(display.redraw)
        return self._form

    def wireup(self):
        super(structure, self).wireup()
        self.form
        self.setConnected(True)

        self.LatticeChoice.sigValueChanged.connect(self.changeUnitCellType)
        self.dimensions.sigValueChanged.connect(self.changeDimensions)

    def setConnected(self, connect):

        for item in [self.LatticeA, self.LatticeB, self.LatticeC]:
            if connect:
                #item.sigValueChanged.connect(self.showUnitCell)
                item.sigValueChanged.connect(self.switchtoCustom)
            else:
                #item.sigValueChanged.disconnect(self.showUnitCell)
                item.sigValueChanged.disconnect(self.switchtoCustom)

    # def showUnitCell(self):
    #     display.redraw()


    def switchtoCustom(self):
        self.LatticeChoice.setValue(0)

    def changeDimensions(self, _, choice):
        self.resetDimensions()

        if choice == '1D':
            self.xspacing.show()
            self.non3DType.show()
            self.domainsize.show()
            self.non3Dlattice.show()
        elif choice == '2D':
            self.xspacing.show()
            self.yspacing.show()
            self.non3DType.show()
            self.domainsize.show()
            self.non3Dlattice.show()
        elif choice == '3D':
            self.Lattice.show()
            self.Basis.show()
            self.Scaling.show()
            self.Repetition.show()

        #self.showUnitCell()

    def resetDimensions(self):
        self.domainsize.hide()
        self.non3DType.hide()
        self.xspacing.hide()
        self.yspacing.hide()
        self.Lattice.hide()
        self.Basis.hide()
        self.Scaling.hide()
        self.Repetition.hide()
        self.non3Dlattice.show()

    def changeUnitCellType(self, _, choice):

        self.setConnected(False)

        if choice == 'Simple Cubic':
            self.LatticeA.setValue((30, 0, 0))
            self.LatticeB.setValue((0, 30, 0))
            self.LatticeC.setValue((0, 0, 30))

        elif choice == 'Body Centered Cubic':
            self.LatticeA.setValue((-30, 30, 15))
            self.LatticeB.setValue((30, 30, 30))
            self.LatticeC.setValue((-30, -30, 30))

        elif choice == 'Face Centered Cubic':
            self.LatticeA.setValue((30, 30, 0))
            self.LatticeB.setValue((0, 30, 30))
            self.LatticeC.setValue((30, 0, 30))

        elif choice == 'Hexagonal Planar':
            self.LatticeA.setValue((10*3, 0, 0))
            self.LatticeB.setValue((5*3, 3 ** .5 * .5*10.*3, 0))
            self.LatticeC.setValue((0, 0, 10*3))

        # TODO: Add HCP

        #self.showUnitCell()

        self.setConnected(True)

    def toStructureDict(self):

        if len(self.parent.ensemble.toDict()):
            ensemble = [('maxgrains', [360, 1, 1])]
            ensemble.append(('orientations', self.parent.ensemble.toDict()))
        else:
            ensemble = [('maxgrains', [1, 1, 1])]

        return UnsortableOrderedDict([('key', 'st' + self.parent.name),
                                      ('iratio', self.iratio.value()),
                                      ('transvec', list(self.Position.value())),
                                      ('grain', self.toGrainDict()),
                                      ('ensemble', UnsortableOrderedDict(ensemble))])

    def toUnitCellDict(self):
        return UnsortableOrderedDict([('key', 'u' + self.parent.name),
                                      ('elements', [UnsortableOrderedDict([('shape_key', self.parent.name),
                                                                           ('locations', self.Basis.toArray())])])])

    def toGrainDict(self):
        z = featuremanager.features.index(self.parent)-1
        # get next lower layer
        while type(featuremanager.features[z]) is not layer and z>=0:
            z-=1
        if type(featuremanager.features[z]) is layer:
            layer_key=featuremanager.features[z].name
        else:
            layer_key=''
        grain = [('unitcell_key', 'u' + self.parent.name),
                 ('layer_key', layer_key)]  # IMPORTANT! How should this be determined?...
        if self.dimensions.value() == '1D':
            grain.extend([('type', self.non3DType.value().lower()),
                          ('lattice',self.non3Dlattice.value().lower()),
                          ('xspacing', UnsortableOrderedDict([('mean',self.xmean.value()),
                                                              ('std',self.xstd.value())])),
                          ('domain', self.domainsize.value())])  # Where does hex/cubic go?...
        elif self.dimensions.value() == '2D':
            grain.extend([('type', self.non3DType.value().lower()),
                          ('lattice',self.non3Dlattice.value().lower()),
                          ('xspacing', UnsortableOrderedDict([('mean',self.xmean.value()),
                                                              ('std',self.xstd.value())])),
                          ('yspacing', UnsortableOrderedDict([('mean',self.ymean.value()),
                                                              ('std',self.ystd.value())])),
                          ('domain', self.domainsize.value())])
        else:
            grain.extend([('lattice', UnsortableOrderedDict([('a', list(self.LatticeA.value())),
                                                             ('b', list(self.LatticeB.value())),
                                                             ('c', list(self.LatticeC.value())),
                                                             # ('basis', self.Basis.toDict()),
                                                             ])),
                          ('scaling', self.Scaling.value()),
                          ('repetition', list(self.Repetition.value()))
                          ])
        return UnsortableOrderedDict(grain)


def loadform(path):
    guiloader = QUiLoader()
    f = QtCore.QFile(path)
    f.open(QtCore.QFile.ReadOnly)
    form = guiloader.load(f)
    f.close()
    return form


class scattering(form):
    def __init__(self, name='Scattering'):
        super(scattering, self).__init__(name)

    @property
    def form(self):
        if self._form is None:
            self.Experiment = pTypes.ListParameter(name='Experiment', values=['GISAXS', 'SAXS'], value=0)
            self.Energy = pTypes.SimpleParameter(name='Photon Energy', type='int', value=10000, suffix='eV',
                                                 siPrefix=True, step=100)
            self.Incidence = StepParameter(name='Incidence Angle(s)', min=.1, max=0, step=0)

            # self.Rotation = StepParameter(name='In-plane Rotation')
            # self.Tilt = StepParameter(name='Tilt', suffix='deg')

            self.IncidenceAngleVisual = pTypes.SimpleParameter(name='Incidence angle', value=0.120, suffix=' deg',
                                                               type='float')
            # self.TiltAngleVisual = pTypes.SimpleParameter(name='Tilt Angle', value=0, suffix=' deg', type='float')
            self.Visualization = pTypes.GroupParameter(name='Beam Visualization (display only)',
                                                       children=[self.IncidenceAngleVisual, self.TiltAngleVisual])

            params = [self.Experiment, self.Energy, self.Incidence, self.Visualization]

            self.parameter = Parameter.create(name='params', type='group', children=params)

            self.parameterTree = ParameterTree()
            self.parameterTree.setParameters(self.parameter, showTop=False)
            self._form = self.parameterTree
        return self._form

    def toDict(self):
        return UnsortableOrderedDict([('pathprefix', '.'),
                                      ('runname', 'BLAH'),
                                      ('expt', self.Experiment.value().lower()),
                                      ('alphai', self.Incidence.toDict()),
                                      ('photon', UnsortableOrderedDict([('value', self.Energy.value()),
                                                                        ('unit', 'ev')])),
                                      ('detector', UnsortableOrderedDict([('pixelsize', 0.172),
                                                                          ('sdd', 4128.62),
                                                                          ('directbeam', [489.916, 843.076])]))])


class detector(form):
    def __init__(self, name='Computation'):
        super(detector, self).__init__(name)

    @property
    def form(self):
        if self._form is None:
            detectornames = ['Custom...'] + [detector().name for detector in detectors.ALL_DETECTORS.values()]
            self.DetectorChoice = pTypes.ListParameter(name='Detector Model', values=detectornames, value='Custom...')
            self.DetectorChoice.sigValueChanged.connect(self.changeDetector)

            self.Width = pTypes.SimpleParameter(name='Width', type='int', suffix=' px', value=300)
            self.Height = pTypes.SimpleParameter(name='Height', type='int', suffix=' px', value=300)
            self.DetectorResolution = pTypes.GroupParameter(name='Detector Resolution',
                                                            children=[self.Width, self.Height])
            self.setConnected(True)

            self.Qparallel = MinMaxParameter(name='Q Parallel', min=-2, max=2)
            self.Qz = MinMaxParameter(name='Q Z', min=0, max=4)
            self.Qrange = pTypes.GroupParameter(name='Q Range', children=[self.Qparallel, self.Qz])

            self.smearing = pTypes.SimpleParameter(name='Smearing', value=0, type='float')

            self.Experiment = pTypes.ListParameter(name='Experiment', values=['GISAXS', 'SAXS'], value=0)
            self.Energy = pTypes.SimpleParameter(name='Photon Energy', type='int', value=10000, suffix='eV',
                                                 siPrefix=True, step=100)
            self.Incidence = StepParameter(name='Incidence Angle(s)', min=.1, max=0, step=0)

            params = [self.DetectorResolution, self.Qrange, self.smearing, self.Experiment, self.Incidence]

            self.parameter = Parameter.create(name='params', type='group', children=params)

            self.parameterTree = ParameterTree()
            self.parameterTree.setParameters(self.parameter, showTop=False)
            self._form = self.parameterTree
        return self._form

    def setConnected(self, connect):
        if connect:
            self.Width.sigValueChanged.connect(self.shapeChanged)
            self.Height.sigValueChanged.connect(self.shapeChanged)
        else:
            self.Width.sigValueChanged.disconnect(self.shapeChanged)
            self.Height.sigValueChanged.disconnect(self.shapeChanged)

    def changeDetector(self, _, choice):
        if not choice == 'Custom...':
            self.setConnected(False)
            detectornames = [detector().name for detector in detectors.ALL_DETECTORS.values()]
            detectorindex = detectornames.index(choice)
            detector = detectors.ALL_DETECTORS.values()[detectorindex]()
            self.Width.setValue(detector.MAX_SHAPE[1])
            self.Height.setValue(detector.MAX_SHAPE[0])
            self.setConnected(True)

    def shapeChanged(self):
        self.DetectorChoice.setValue('Custom...')

    def toDict(self):
        return UnsortableOrderedDict([('path', '.'),
                                      ('runname', 'BLAH'),
                                      ('expt', self.Experiment.value().lower()),
                                      ('alphai', self.Incidence.toDict()),
                                      ('photon', UnsortableOrderedDict([('energy', self.Energy.value()),
                                                                        ('unit', 'ev')])),
                                      ('smearing', 0),
                                      ('detector', self.DetectorChoice.value()),
                                      ('output', UnsortableOrderedDict([('type', 'qspace'),
                                                                        ('minpoint', [self.Qparallel.value()[0],
                                                                                      self.Qz.value()[0]]),
                                                                        ('maxpoint', [self.Qparallel.value()[1],
                                                                                      self.Qz.value()[1]])])),
                                      ('resolution', [self.Width.value(), self.Height.value()])
                                      ])


class VectorParameterItem(pTypes.WidgetParameterItem):
    def makeWidget(self, cls = vector):
        w = cls()
        opts = self.param.opts
        value = opts.get('value', None)
        if value is not None:
            w.setValue(value)
        else:
            w.setValue([0,0,0])

        self.value = w.value
        self.setValue = w.setValue

        return w

class IntVectorParameterItem(VectorParameterItem):
    def makeWidget(self):
        return super(IntVectorParameterItem, self).makeWidget(cls=intvector)


class VectorParameter(Parameter):
    itemClass = VectorParameterItem

    def __init__(self, *args, **kwargs):
        super(VectorParameter, self).__init__(*args, **kwargs)

    def defaultValue(self):
        return (0, 0, 0)

class IntVectorParameter(VectorParameter):
    itemClass = IntVectorParameterItem

registerParameterType('Vector', VectorParameter, override=True)


class hideableGroupParameterItem(pTypes.GroupParameterItem):
    def optsChanged(self, param, opts):
        super(hideableGroupParameterItem, self).optsChanged(param, opts)
        if 'visible' in opts:
            self.setHidden(not opts['visible'])


class hideableGroupParameter(pTypes.GroupParameter):
    itemClass = hideableGroupParameterItem


class ScalableGroup(hideableGroupParameter):
    def __init__(self, **opts):
        opts['type'] = 'group'
        opts['addText'] = "Add"
        pTypes.GroupParameter.__init__(self, **opts)

    def addNew(self):
        self.addChild(dict(name="Point %d" % (len(self.childs) + 1), type='Vector', removable=True, renamable=True, value=[0,0,0]))

    def toArray(self):
        return [list(child.value()) for child in self.children()]

    def toDict(self):
        d = dict()
        for child in self.children():
            d[child.opts['name']] = list(child.value())
        return d


class RotationDistParameter(pTypes.GroupParameter):
    itemClass = hideableGroupParameterItem

    def __init__(self, value=None, **opts):
        opts['type'] = 'bool'
        opts['value'] = True
        pTypes.GroupParameter.__init__(self, **opts)

        self.DistributionChoice = pTypes.ListParameter(name='Distribution', type='list', value=0,
                                                       values=['Uniform', 'Random', 'Gaussian', 'Bragg'])
        self.Min = pTypes.SimpleParameter(name='Minimum', type='float', value=0)
        self.Max = pTypes.SimpleParameter(name='Maximum', type='float', value=0)
        self.Mean = pTypes.SimpleParameter(name='Mean', type='float', value=0)
        self.Variance = pTypes.SimpleParameter(name='Variance', type='float', value=0)

        self.Mean.hide()
        self.Variance.hide()

        self.DistributionChoice.sigValueChanged.connect(self.distributionChanged)

        self.addChildren([self.DistributionChoice, self.Min, self.Max, self.Mean, self.Variance])

    def distributionChanged(self, _, choice):
        # print choice
        if choice == 'Uniform':
            self.Min.show()
            self.Max.show()
            self.Mean.hide()
            self.Variance.hide()
        elif choice == 'Random':
            self.Min.show()
            self.Max.show()
            self.Mean.hide()
            self.Variance.hide()
        elif choice == 'Gaussian':
            self.Min.show()
            self.Max.show()
            self.Mean.show()
            self.Variance.show()
        elif choice == 'Bragg':
            self.Min.hide()
            self.Max.hide()
            self.Mean.hide()
            self.Variance.hide()

    def toDict(self):
        d = UnsortableOrderedDict()
        choice = self.DistributionChoice.value()

        if choice == 'Uniform':
            d['min'] = self.Min.value()
            d['max'] = self.Max.value()
            d['stat'] = 'uniform'
        elif choice == 'Random':
            d['min'] = self.Min.value()
            d['max'] = self.Max.value()
            d['stat'] = 'random'
        elif choice == 'Gaussian':
            d['min'] = self.Min.value()
            d['max'] = self.Max.value()
            d['std'] = float(np.sqrt(self.Variance.value()))
            d['mean'] = self.Mean.value()
            d['stat'] = 'gaussian'
        return d


class DistParameter(pTypes.GroupParameter):
    itemClass = hideableGroupParameterItem

    def __init__(self, value=None, **opts):
        opts['type'] = 'bool'
        opts['value'] = True
        pTypes.GroupParameter.__init__(self, **opts)

        self.DistributionChoice = pTypes.ListParameter(name='Distribution', type='list', value=0,
                                                       values=['Single value', 'Uniform', 'Random', 'Gaussian'])
        self.Value = pTypes.SimpleParameter(name=opts['name'], type='float', value=value)
        self.Min = pTypes.SimpleParameter(name='Minimum', type='float', value=0)
        self.Max = pTypes.SimpleParameter(name='Maximum', type='float', value=0)
        self.Mean = pTypes.SimpleParameter(name='Mean', type='float', value=0)
        self.Variance = pTypes.SimpleParameter(name='Variance', type='float', value=0)
        self.N = pTypes.SimpleParameter(name='Number of samples', type='int', value=0)

        self.Min.hide()
        self.Max.hide()
        self.Mean.hide()
        self.Variance.hide()
        self.N.hide()

        self.DistributionChoice.sigValueChanged.connect(self.distributionChanged)

        self.addChildren([self.DistributionChoice, self.Value, self.Min, self.Max, self.Mean, self.Variance, self.N])

    def distributionChanged(self, _, choice):
        # print choice
        if choice == 'Uniform':
            self.Value.hide()
            self.Min.show()
            self.Max.show()
            self.Mean.hide()
            self.Variance.hide()
            self.N.show()
        elif choice == 'Single value':
            self.Value.show()
            self.Min.hide()
            self.Max.hide()
            self.Mean.hide()
            self.Variance.hide()
            self.N.hide()
        elif choice == 'Random':
            self.Value.hide()
            self.Min.show()
            self.Max.show()
            self.Mean.hide()
            self.Variance.hide()
            self.N.show()
        elif choice == 'Gaussian':
            self.Value.hide()
            self.Min.show()
            self.Max.show()
            self.Mean.show()
            self.Variance.show()
            self.N.show()

    def toDict(self):
        d = UnsortableOrderedDict()
        choice = self.DistributionChoice.value()

        if 'higKey' in self.opts:
            d['type'] = self.opts['higKey'].lower()

        if choice == 'Uniform':
            d['min'] = self.Min.value()
            d['max'] = self.Max.value()
            d['N'] = self.N.value()
            d['stat'] = 'uniform'
        elif choice == 'Single value':
            d['min'] = self.Value.value()
            d['stat'] = 'single'
        elif choice == 'Random':
            d['min'] = self.Min.value()
            d['max'] = self.Max.value()
            d['N'] = self.N.value()
            d['stat'] = 'random'
        elif choice == 'Gaussian':
            d['min'] = self.Min.value()
            d['max'] = self.Max.value()
            d['N'] = self.N.value()
            d['std'] = np.sqrt(self.Variance.value())
            d['mean'] = self.Mean.value()
            d['stat'] = 'gaussian'
        return d


class RotationParameter(pTypes.GroupParameter):
    def __init__(self, **opts):
        opts['type'] = 'bool'
        opts['value'] = True
        super(RotationParameter, self).__init__(**opts)

        self.AxisChoice = pTypes.ListParameter(name='Axis', type='list', value=0,
                                               values=['X', 'Y', 'Z'])
        self.AngleDist = RotationDistParameter(name='Angles')
        self.delete = pTypes.ActionParameter(name='Remove Rotation')
        self.delete.sigActivated.connect(self.remove)  # Will this work?

        self.addChildren([self.AxisChoice, self.AngleDist, self.delete])

    def toDict(self):
        d = self.AngleDist.toDict()
        d['axis'] = self.AxisChoice.value().lower()
        return d


class StepParameter(pTypes.GroupParameter):
    itemClass = hideableGroupParameterItem

    def __init__(self, min=None, max=None, step=None, **opts):
        opts['type'] = 'bool'
        opts['value'] = True
        pTypes.GroupParameter.__init__(self, **opts)

        self.Min = pTypes.SimpleParameter(name='Minimum', type='float', value=min)
        self.Max = pTypes.SimpleParameter(name='Maximum', type='float', value=max)
        self.Step = pTypes.SimpleParameter(name='Step', type='float', value=step)

        self.addChildren([self.Min, self.Max, self.Step])

    def toDict(self):
        d = dict()
        d['min'] = self.Min.value()
        d['max'] = self.Max.value()
        d['step'] = self.Step.value()

        return d


class MinMaxParameter(pTypes.GroupParameter):
    itemClass = hideableGroupParameterItem

    def __init__(self, min=None, max=None, **opts):
        opts['type'] = 'bool'
        opts['value'] = True
        pTypes.GroupParameter.__init__(self, **opts)

        self.Min = pTypes.SimpleParameter(name='Minimum', type='float', value=min)
        self.Max = pTypes.SimpleParameter(name='Maximum', type='float', value=max)

        self.addChildren([self.Min, self.Max])

    def value(self):
        return [self.Min.value(), self.Max.value()]

    def toDict(self):
        d = UnsortableOrderedDict()
        d['min'] = self.Min.value()
        d['max'] = self.Max.value()

        return d



import importlib
from pipeline import msg
from PySide import QtGui

def import_module(modname,packagename=None):
    module=None
    try:
        module=importlib.import_module(modname, packagename)
        msg.logMessage(("Imported", modname), msg.DEBUG)
    except ImportError as ex:
        msg.logMessage('Module could not be loaded: ' + modname)

        missingpackage = ex.message.replace('No module named ', '')

        import config
        if missingpackage in ['qt.qtapi','pandas','tomopy','matplotlib.pyplot']:
            return None

        if config.settings['Ignored Modules']:
            if missingpackage in config.settings['Ignored Modules']:
                return None

        msgBox = QtGui.QMessageBox()
        msgBox.setText("A python package is missing! Xi-cam can try to install this for you.")
        msgBox.setInformativeText("Would you like to install " + missingpackage + "?")
        msgBox.setStandardButtons(QtGui.QMessageBox.Yes | QtGui.QMessageBox.No | QtGui.QMessageBox.Ignore)
        msgBox.setDefaultButton(QtGui.QMessageBox.Yes)

        response = msgBox.exec_()

        if response == QtGui.QMessageBox.Yes:
            import pip
            if not hasattr(pip, 'main'):
                import pip._internal as pip

            failure=pip.main(['install', '--user', missingpackage])
            if failure:
                failure=pip.main(['install', missingpackage])

            if not failure:
                msgBox = QtGui.QMessageBox()
                msgBox.setText('Success! The missing package, ' + missingpackage + ', has been installed!')
                msgBox.setInformativeText('Please restart Xi-cam now.')
                msgBox.setStandardButtons(QtGui.QMessageBox.Ok)
                msgBox.exec_()
                exit(0)
            else:
                if modname.strip('.') == 'MOTD':
                    from xicam import debugtools
                    debugtools.frustration()
                    msgBox = QtGui.QMessageBox()
                    msgBox.setText(
                        'Sorry, ' + missingpackage + ' could not be installed. This is a Xi-cam critical library.')
                    msgBox.setInformativeText('Xi-cam cannot be loaded . Please install ' + missingpackage + ' manually.')
                    msgBox.setStandardButtons(QtGui.QMessageBox.Ok)
                    msgBox.exec_()
                    exit(1)
                else:
                    from xicam import debugtools
                    debugtools.frustration()
                    msgBox = QtGui.QMessageBox()
                    msgBox.setText(
                        'Sorry, ' + missingpackage + ' could not be installed. Try installing this package yourself, or contact the package developer.')
                    msgBox.setInformativeText('Would you like to continue loading Xi-cam?')
                    msgBox.setStandardButtons(QtGui.QMessageBox.Yes | QtGui.QMessageBox.No)
                    response = msgBox.exec_()
                    if response == QtGui.QMessageBox.No:
                        exit(1)
        elif response == QtGui.QMessageBox.Ignore and modname.strip('.') != 'MOTD':
            import config
            if config.settings['Ignored Modules']:
                config.settings['Ignored Modules']+=[missingpackage]
            else:
                config.settings['Ignored Modules']=[missingpackage]
            msgBox = QtGui.QMessageBox()
            msgBox.setText('Xi-cam will no longer prompt you to install this package, however some plugins may be disabled.')
            msgBox.setStandardButtons(QtGui.QMessageBox.Ok)
            msgBox.exec_()

        if modname.strip('.') == 'MOTD':
            from xicam import debugtools
            debugtools.frustration()
            msgBox = QtGui.QMessageBox()
            msgBox.setText(
                'Sorry, ' + missingpackage + ' is a Xi-cam critical library. This must be installed to run Xi-cam!')
            msgBox.setInformativeText('Xi-cam cannot be loaded . Please install ' + modname.strip('.') + ' manually.')
            msgBox.setStandardButtons(QtGui.QMessageBox.Ok)
            msgBox.exec_()
            exit(1)


    if not module: msg.logMessage('Failed to import '+modname,msg.CRITICAL)
    return module


import platform
from fabio import edfimage
from pipeline import msg

# Use NSURL as a workaround to pyside/Qt4 behaviour for dragging and dropping on OSx
op_sys = platform.system()
if op_sys == 'Darwin':
    try:
        from Foundation import NSURL
    except ImportError:
        msg.logMessage('NSURL not found. Drag and drop may not work correctly',msg.WARNING)

import base
from PySide import QtGui, QtCore
import os
from pyqtgraph.parametertree import ParameterTree
import widgets
import numpy as np
from pipeline.spacegroups import spacegroupwidget
from pipeline import loader
from xicam import config
import fabio
from pipeline import calibration

from xicam.widgets.calibrationpanel import calibrationpanel

# Globals so Timeline can share the same rightmodes
configtree = ParameterTree()
configtree.setParameters(config.activeExperiment, showTop=False)


def tiltStyleMenuRequested(pos):
    config.activeExperiment.tiltStyleMenu.exec_(configtree.mapToGlobal(pos))


configtree.customContextMenuRequested.connect(tiltStyleMenuRequested)
configtree.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)




rightmodes = [(configtree, QtGui.QFileIconProvider().icon(QtGui.QFileIconProvider.File))]

class ViewerPlugin(base.plugin):
    name = 'Viewer'
    sigUpdateExperiment = QtCore.Signal()


    def __init__(self, *args, **kwargs):

        self.centerwidget = QtGui.QTabWidget()
        self.centerwidget.currentChanged.connect(self.currentChanged)
        self.centerwidget.setDocumentMode(True)
        self.centerwidget.setTabsClosable(True)
        self.centerwidget.tabCloseRequested.connect(self.tabCloseRequested)

        self.rightmodes = rightmodes
        self.bottomwidget = widgets.integrationwidget(self.getCurrentTab)



        self.toolbar = widgets.toolbar.difftoolbar()
        self.toolbar.connecttriggers(self.calibrate, self.centerfind, self.refinecenter, self.redrawcurrent,
                                     self.redrawcurrent, self.remeshmode, self.linecut, self.vertcut,
                                     self.horzcut, self.redrawcurrent, self.redrawcurrent, self.redrawcurrent,
                                     self.roi, self.arccut, self.polymask,
                                     capture=None, removecosmics=self.removecosmics,thresholdmask=self.thresholdmask)


        self.spacegroupwidget = spacegroupwidget()
        self.spacegroupwidget.sigDrawSGOverlay.connect(self.drawsgoverlay)
        sgicon = QtGui.QIcon()
        sgicon.addPixmap(QtGui.QPixmap("xicam/gui/icons_35.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.rightmodes.append((self.spacegroupwidget,sgicon))

        self.propertytable = widgets.frameproptable()
        self.rightmodes.append((self.propertytable,QtGui.QFileIconProvider().icon(QtGui.QFileIconProvider.Desktop)))

        self.calibrationPanel = calibrationpanel()
        calicon = QtGui.QIcon()
        calicon.addPixmap(QtGui.QPixmap("xicam/gui/icons_28.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.rightmodes.append((self.calibrationPanel, calicon))
        self.calibrationPanel.sigCalibrate.connect(self.calibrate)
        self.calibrationPanel.sigSimulateCalibrant.connect(self.simulatecalibrant)

        super(ViewerPlugin, self).__init__(*args, **kwargs)

        config.activeExperiment.sigTreeStateChanged.connect(self.sigUpdateExperiment)
        self.sigUpdateExperiment.connect(self.redrawcurrent)
        self.sigUpdateExperiment.connect(self.replotcurrent)
        self.sigUpdateExperiment.connect(self.invalidatecache)

        self.booltoolbar.actionAdd.triggered.connect(self.addmode)
        self.booltoolbar.actionSubtract.triggered.connect(self.subtractmode)
        self.booltoolbar.actionAdd_with_coefficient.triggered.connect(self.addwithcoefmode)
        self.booltoolbar.actionSubtract_with_coefficient.triggered.connect(self.subtractwithcoefmode)
        self.booltoolbar.actionDivide.triggered.connect(self.dividemode)
        self.booltoolbar.actionAverage.triggered.connect(self.averagemode)



        # DRAG-DROP
        self.centerwidget.setAcceptDrops(True)
        self.centerwidget.dragEnterEvent = self.dragEnterEvent
        self.centerwidget.dropEvent = self.dropEvent


    def dragEnterEvent(self, e):
        e.accept()
        # if e.mimeData().hasFormat('text/plain'):
        # e.accept()
        # else:
        #     e.accept()

    def dropEvent(self, e):
        for url in e.mimeData().urls():
            if op_sys == 'Darwin':
                fname = str(NSURL.URLWithString_(str(url.toString())).filePathURL().path())
            else:
                fname = str(url.toLocalFile())
            if os.path.isfile(fname):
                msg.logMessage(fname,msg.DEBUG)
                self.openfiles([fname])
            e.accept()




    def drawsgoverlay(self, peakoverlay):
        self.getCurrentTab().drawsgoverlay(peakoverlay)

    def simulatecalibrant(self,calibrant):
        self.getCurrentTab().simulatecalibrant(calibrant)

    def addmode(self):
        """
        Launch a tab as an add operation
        """
        operation = lambda m: np.sum(m, 0)
        self.launchmultimode(operation, 'Addition')

    def subtractmode(self):
        """
        Launch a tab as an sub operation
        """
        operation = lambda m: m[0] - np.sum(m[1:], 0)
        self.launchmultimode(operation, 'Subtraction')

    def addwithcoefmode(self):
        """
        Launch a tab as an add with coef operation
        """
        coef, ok = QtGui.QInputDialog.getDouble(self.ui, u'Enter scaling coefficient x (A+xB):', u'Enter coefficient')

        if coef and ok:
            operation = lambda m: m[0] + coef * np.sum(m[1:], 0)
            self.launchmultimode(operation, 'Addition with coef (x=' + coef + ')')

    def subtractwithcoefmode(self):
        """
        Launch a tab as a sub with coef operation
        """
        coef, ok = QtGui.QInputDialog.getDouble(None, u'Enter scaling coefficient x (A-xB):', u'Enter coefficient')

        if coef and ok:
            operation = lambda m: m[0] - coef * np.sum(m[1:], 0)
            self.launchmultimode(operation, 'Subtraction with coef (x=' + str(coef))

    def dividemode(self):
        """
        Launch a tab as a div operation
        """
        operation = lambda m: m[0] / m[1]
        self.launchmultimode(operation, 'Division')

    def averagemode(self):
        """
        Launch a tab as an avg operation
        """
        operation = lambda m: np.mean(m, 0)
        self.launchmultimode(operation, 'Average')

    def launchmultimode(self, operation, operationname):
        """
        Launch a tab in multi-image operation mode
        """
        self.openSelected(operation, operationname)

    def tabCloseRequested(self, index):
        self.centerwidget.widget(index).deleteLater()

    def getCurrentTab(self):
        if self.centerwidget.currentWidget() is None: return None
        if not hasattr(self.centerwidget.currentWidget(),'widget'): return None
        return self.centerwidget.currentWidget().widget

    def calibrate(self, algorithm=calibration.fourierAutocorrelation, calibrant='AgBh'):
        self.getCurrentTab().calibrate(algorithm, calibrant)

    def centerfind(self):
        self.getCurrentTab().centerfind()

    def refinecenter(self):
        self.getCurrentTab().refinecenter()

    def redrawcurrent(self):
        currenttab = self.getCurrentTab()
        if currenttab: self.getCurrentTab().redrawimage()


    def remeshmode(self):
        self.getCurrentTab().redrawimage()
        self.getCurrentTab().replot()

    def linecut(self):
        self.getCurrentTab().linecut()

    def vertcut(self):
        self.getCurrentTab().verticalcut()

    def horzcut(self):
        self.getCurrentTab().horizontalcut()

    def roi(self):
        self.getCurrentTab().roi()

    def arccut(self):
        self.getCurrentTab().arccut()

    def polymask(self):
        self.getCurrentTab().polymask()

    def thresholdmask(self):
        self.getCurrentTab().thresholdmask()


    def currentChanged(self, index):
        for tab in [self.centerwidget.widget(i) for i in range(self.centerwidget.count())]:
            tab.unload()
        self.centerwidget.currentWidget().load()
        self.propertytable.setData(self.currentImage().dimg.headers)
        self.replotcurrent()


    def openfiles(self, paths=None, operation=None, operationname=None):
        self.activate()
        if type(paths) is not list:
            paths = [paths]

        widget = widgets.OOMTabItem(itemclass=widgets.dimgViewer, src=paths, operation=operation,
                                    operationname=operationname, plotwidget=self.bottomwidget,
                                    toolbar=self.toolbar)
        self.centerwidget.addTab(widget, os.path.basename(paths[0]))
        self.centerwidget.setCurrentWidget(widget)

    def opendata(self, data=None, operation=None, operationname=None):
        self.activate()
        dimg = loader.datadiffimage2(data=data)
        widget = widgets.OOMTabItem(itemclass=widgets.dimgViewer, dimg=dimg, operation=operation,
                                    operationname=operationname, plotwidget=self.bottomwidget,
                                    toolbar=self.toolbar)
        self.centerwidget.addTab(widget, 'Imported data')
        self.centerwidget.setCurrentWidget(widget)

    def currentImage(self):
        return self.getCurrentTab()

    def replotcurrent(self):
        if self.getCurrentTab(): self.getCurrentTab().replot()

    def invalidatecache(self):
        if self.getCurrentTab(): self.getCurrentTab().dimg.invalidatecache()

    def togglespacegroup(self):
        if self.toolbar.actionSpaceGroup.isChecked():
            self.placeholders[1].setCurrentWidget(self.spacegroupwidget)
        else:
            self.placeholders[1].setCurrentWidget(self.rightwidget)

    def exportimage(self):
        fabimg = edfimage.edfimage(np.rot90(self.getCurrentTab().imageitem.image))
        dialog = QtGui.QFileDialog(parent=None, caption=u"Export image as EDF",
                                   directory=unicode(os.path.dirname(self.getCurrentTab().dimg.filepath)),
                                   filter=u"EDF (*.edf)")
        dialog.selectFile(unicode(os.path.dirname(self.getCurrentTab().dimg.filepath)))
        filename, ok = dialog.getSaveFileName()
        msg.logMessage(filename,msg.DEBUG)
        if ok and filename:
            fabimg.write(filename)

    def capture(self):
        self.getCurrentTab().capture()

    def maskload(self):
        filename, ok = QtGui.QFileDialog.getOpenFileName(None, 'Load Mask', os.curdir)

        if filename and ok:
            self.openfiles([filename])
        config.activeExperiment.addtomask(np.rot90(fabio.open(filename).data,3))

    def removecosmics(self):
        self.getCurrentTab().removecosmics()




__author__ = "Luis Barroso-Luque"
__copyright__ = "Copyright 2016, CAMERA, LBL, ALS"
__credits__ = ["Ronald J Pandolfi", "Dinesh Kumar", "Singanallur Venkatakrishnan", "Luis Luque", "Alexander Hexemer"]
__license__ = ""
__version__ = "1.2.1"
__maintainer__ = "Ronald J Pandolfi"
__email__ = "ronpandolfi@lbl.gov"
__status__ = "Beta"


import threads
import client
from pipeline import msg

# Some default HPC host addresses
HPC_SYSTEM_ADDRESSES = {'Cori': 'cori.nersc.gov', 'Edison': 'edison.nersc.gov', 'Bragg': 'bragg.dhcp.lbl.gov'}

# Clients and what not
# bind classes to new names
sftp_client = client.sftp.SFTPClient
globus_client = client.globus.GlobusClient
ssh_client = client.ssh.SSHClient

# Singleton instance of spot_client
spot_client = client.spot.SpotClient()

# Dicts to hold client instances
sftp_clients = {}
globus_clients = {}
ssh_clients = {}


def login_wrapper(client_login):
    """Decorator to catch all login errors from NEWT, Globus, and PySFTP/Paramiko"""
    def handled_login(*args, **kwargs):
        try:
            return client_login(*args, **kwargs)
        except client.EXCEPTIONS as e:
            msg.logMessage(e.message,msg.ERROR)
            return

    return handled_login


def login(client_callback, client_login, credentials):
    """Login clients on a background thread"""
    handled_login = login_wrapper(client_login)
    bg_handled_login = threads.method(callback_slot=client_callback)(handled_login)
    bg_handled_login(**credentials)


def add_sftp_client(host, client, callback):
    """Add sftp client to dictionary in order to have them accessible to plugins"""
    sftp_clients[host] = client
    callback(client)


def add_ssh_client(host, client, callback):
    """Add sftp client to dictionary in order to have them accessible to plugins"""
    ssh_clients[host] = client
    callback(client)


def add_globus_client(endpoint, client, callback):
    """Add globus client to dictionary in order to have them accessible to plugins"""
    globus_clients[endpoint] = client
    callback(client)


def logout(client_obj, callback=None):
    """Logout client on a background thread"""
    # TODO remove clients from their corresponding dictionaries!!!
    if hasattr(client_obj, 'logout'):
        method = client_obj.logout
    elif hasattr(client_obj, 'close'):
        method = client_obj.close
    threads.method(callback_slot=callback)(method)()  # Decorate and run client logout/close method


# TODO implement this to save NIM credentials
class NIMCredentials(object):
    """Class to save NIM user credentials to avoid inputting them soooo many times"""
    # Is this not secure? I am mangling the names though...
    def __init__(self, user, password):
        self.__user = user
        self.__password = password



__author__ = 'remi'
################
import sys
from numpy import NaN, Inf, arange, isscalar, asarray, array
import pyqtgraph as pg


def peakdet(x, v, delta):
    maxtab = []
    mintab = []

    if x is None:
        x = arange(len(v))

    v = asarray(v)

    if len(v) != len(x):
        sys.exit('Input vectors v and x must have same length')

    if not isscalar(delta):
        sys.exit('Input argument delta must be a scalar')

    if delta <= 0:
        sys.exit('Input argument delta must be positive')

    mn, mx = Inf, -Inf
    mnpos, mxpos = NaN, NaN

    lookformax = True

    for i in arange(len(v)):
        this = v[i]
        if this > mx:
            mx = this
            mxpos = x[i]
        if this < mn:
            mn = this
            mnpos = x[i]

        if lookformax:
            if this < mx - delta:
                maxtab.append((mxpos, mx))
                mn = this
                mnpos = x[i]
                lookformax = False
        else:
            if this > mn + delta:
                mintab.append((mnpos, mn))
                mx = this
                mxpos = x[i]
                lookformax = True

    return array(maxtab), array(mintab)


###############

class peaktooltip:
    def __init__(self, x, y, widget):
        self.q, self.I, self.width, self.index = peakdet(x, y, 10)
        self.scatterPoints = pg.ScatterPlotItem(self.q, self.I, size=10, pen=pg.mkPen(None),
                                                brush=pg.mkBrush(255, 255, 255, 120))
        self.display_text = pg.TextItem(text='', color=(176, 23, 31), anchor=(0, 1))
        self.display_text.hide()
        widget.addItem(self.scatterPoints)
        widget.addItem(self.display_text)
        self.scatterPoints.scene().sigMouseMoved.connect(self.onMove)

    def onMove(self, pos):
        act_pos = self.scatterPoints.mapFromScene(pos)
        p1 = self.scatterPoints.pointsAt(act_pos)
        """
        :type p1 : pg.graphicsItems.ScatterPlotItem.SpotItem
        """
        if len(p1) != 0:
            self.display_text.setText('q=%f\nI=%f' % (p1[0].pos().x(), p1[0].pos().y()))
            self.display_text.setPos(p1[0].pos())
            self.display_text.show()
        else:
            self.display_text.hide()


from collections import OrderedDict
from PySide import QtGui
import sys
from xicam import xglobals
import importlib
from pipeline import msg
import os
import pkgutil
from xicam import safeimporter

import inspect

modules = []
plugins = OrderedDict()

disabledatstart = ['FXS', 'SPOTH5', 'Library', 'XAS','EZTest', '3D Viewer', 'HipRMC', ]


def initplugins(placeholders):
    import base
    global plugins, modules

    packages = pkgutil.iter_modules(__path__)
    msg.logMessage(('packages:',packages),msg.DEBUG)

    for importer, modname, ispkg in packages:

        msg.logMessage("Found plugin %s (is a package: %s)" % (modname, ispkg),msg.DEBUG)
        mod=safeimporter.import_module('.' + modname, 'xicam.plugins')
        if mod:
            modules.append(mod)
        else:
            msg.logMessage("Plugin loading aborted: "+modname,msg.CRITICAL)
            continue


    for module in modules:
        msg.logMessage(('Imported, enabling:',module.__name__),msg.DEBUG)
        for objname,obj in module.__dict__.copy().iteritems():
            if inspect.isclass(obj):
                if issubclass(obj, base.plugin) and not obj is base.plugin and not obj is base.EZplugin:
                    link = pluginlink(module, obj, placeholders)
                    if link.name not in disabledatstart: link.enable()
                    plugins[link.name] = link
            elif type(obj) is base.EZplugin:
                link = pluginlink(module, obj, placeholders)
                if link.name not in disabledatstart: link.enable()
                plugins[link.name] = link

    xglobals.plugins = plugins


def buildactivatemenu(modewidget):
    menu = QtGui.QMenu('Plugins')
    for pluginlink in plugins.values():
        if pluginlink.plugin.hidden:
            continue
        action = QtGui.QAction(pluginlink.name, menu)
        action.setCheckable(True)
        action.setChecked(pluginlink.enabled)
        action.toggled.connect(pluginlink.toggle)
        action.toggled.connect(modewidget.reload)
        menu.addAction(action)
    return menu


class pluginlink():
    def __init__(self, module, plugin, placeholders):
        self.plugin = plugin
        self.modulename = module.__name__
        self.module = module
        self.instance = None
        self.placeholders = placeholders
        self.name = self.plugin.name
        self._enabled = False

    def disable(self):
        del self.instance
        self.instance = None
        self._enabled = False

    def enable(self):
        #self.module = reload(sys.modules[self.modulename])
        if inspect.isclass(self.plugin):
            self.instance = self.plugin(self.placeholders)
        else:
            self.instance = self.plugin
            self.plugin.setup(self.placeholders)
        self._enabled = True

    @property
    def enabled(self):
        return self._enabled

    @enabled.setter
    def enabled(self, enabled):
        if enabled and not self._enabled:
            self.enable()
        elif not enabled and self._enabled:
            self.disable()

    def toggle(self):
        self.disable() if self.enabled else self.enable()

    def activate(self):
        self.enabled = True
        self.instance.activate()


from PySide import QtGui
from PySide import QtCore
from PySide.QtCore import Qt
from PIL import Image
import viewer
import os
import pipeline
import numpy as np
from pipeline import msg

class FlowLayout(QtGui.QLayout):
    """
    A grid layout that wraps and adjusts to its items
    """
    def __init__(self, parent=None, margin=5, spacing=-1):
        super(FlowLayout, self).__init__(parent)

        self.margin = margin
        self.setSpacing(spacing)

        self.itemList = []


    def __del__(self):
        item = self.takeAt(0)
        while item:
            item = self.takeAt(0)

    def addItem(self, item):
        self.itemList.append(item)

    def count(self):
        return len(self.itemList)

    def itemAt(self, index):
        if index >= 0 and index < len(self.itemList):
            return self.itemList[index]

        return None

    def takeAt(self, index):
        if index >= 0 and index < len(self.itemList):
            return self.itemList.pop(index)

        return None

    def expandingDirections(self):
        return Qt.Orientations(Qt.Orientation(0))

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        height = self.doLayout(QtCore.QRect(0, 0, width, 0), True)
        return height

    def setGeometry(self, rect):
        super(FlowLayout, self).setGeometry(rect)
        self.doLayout(rect, False)

    def sizeHint(self):
        return self.minimumSize()

    def minimumSize(self):
        size = QtCore.QSize()

        for item in self.itemList:
            size = size.expandedTo(item.minimumSize())

        size += QtCore.QSize(2 * self.margin, 2 * self.margin)
        return size

    def doLayout(self, rect, testOnly):
        x = rect.x()
        y = rect.y()
        lineHeight = 0

        for item in self.itemList:
            wid = item.widget()
            spaceX = self.spacing() + wid.style().layoutSpacing(QtGui.QSizePolicy.PushButton,
                                                                QtGui.QSizePolicy.PushButton,
                                                                Qt.Horizontal)
            spaceY = self.spacing() + wid.style().layoutSpacing(QtGui.QSizePolicy.PushButton,
                                                                QtGui.QSizePolicy.PushButton,
                                                                Qt.Vertical)
            nextX = x + item.sizeHint().width() + spaceX
            if nextX - spaceX > rect.right() and lineHeight > 0:
                x = rect.x()
                y = y + lineHeight + spaceY
                nextX = x + item.sizeHint().width() + spaceX
                lineHeight = 0

            if not testOnly:
                item.setGeometry(QtCore.QRect(QtCore.QPoint(x, y), item.sizeHint()))

            x = nextX
            lineHeight = max(lineHeight, item.sizeHint().height())

        return y + lineHeight - rect.y()


class librarylayout(FlowLayout):
    """
    Extend the flow layout to fill it with thumbwidgetitems representing files/folders
    """

    def __init__(self, parentwindow, path='/Volumes'):
        super(librarylayout, self).__init__()
        self.parent = None
        self.parentwindow = parentwindow
        self.populate(path)

    def chdir(self, path):
        self.clear()
        self.populate(os.path.normpath(path))

    def populate(self, path):
        self.parent = QtCore.QDir()
        self.parent.cd(path)

        dir = QtCore.QDir(self.parent)
        entries = dir.entryList()

        for entry in entries:
            # print fileinfo.fileName()
            if not (entry == '..' and os.path.normpath(path) == pipeline.pathtools.getRoot()) and not entry == '.':
                self.addWidget(thumbwidgetitem(os.path.join(path,entry), parentwindow=self.parentwindow))


    def clear(self):
        while self.count():
            item = self.takeAt(0)
            widget = item.widget()
            if widget is not None:
                widget.deleteLater()

class thumbwidgetitem(QtGui.QFrame):
    """
    A single icon representing a file/folder that can be accessed/opened
    """



    def __init__(self, path, parentwindow):
        path = os.path.normpath(path)

        self.foldericon = QtGui.QImage()
        self.foldericon.load('xicam/gui/GenericFolderIcon.png')

        self.fileicon = QtGui.QImage()
        self.fileicon.load('xicam/gui/post-360412-0-09676400-1365986245.png')

        super(thumbwidgetitem, self).__init__()
        self.parentwindow = parentwindow
        self.setObjectName('thumb')
        desiredsize = QtCore.QSize(160, 200)


        self.setFixedSize(desiredsize)
        self.setAutoFillBackground(True)
        self.setFocusPolicy(Qt.StrongFocus)


        self.layout = QtGui.QVBoxLayout(self)  #.frame

        self.path = path
        # print path
        self.dimg = pipeline.loader.diffimage(filepath=self.path)
        self.image = QtGui.QImage()
        #print os.path.splitext(path)[1]
        if os.path.isdir(path):
            self.image = self.foldericon
        elif os.path.splitext(path)[1] in pipeline.loader.acceptableexts:


            try:
                self.thumb = np.rot90(np.log(self.dimg.thumbnail + 1)).copy()
                self.thumb *= 255. / np.max(self.thumb)



                # if self.thumb is not None:
                self.image = QtGui.QImage(self.thumb.astype(np.uint8), self.thumb.shape[1], self.thumb.shape[0],
                                          self.thumb.shape[1],
                                          QtGui.QImage.Format_Indexed8)
            except Exception as ex:
                msg.logMessage(ex.message,msg.Error)

                self.image = self.fileicon


        else:
            self.image = self.fileicon

        image_label = ScaledLabel(self.image)
        image_label.setAlignment(Qt.AlignHCenter)
        self.layout.addWidget(image_label)

        self.namelabel = QtGui.QLabel(os.path.basename(path))
        self.namelabel.setAlignment(Qt.AlignHCenter)
        self.namelabel.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Maximum)

        line = QtGui.QFrame()
        line.setGeometry(QtCore.QRect(0, 0, desiredsize.width() * 9 / 10, 3))
        line.setFrameShape(QtGui.QFrame.HLine)
        line.setFrameShadow(QtGui.QFrame.Sunken)
        self.layout.addWidget(line)

        self.layout.addWidget(self.namelabel)
        self.setLayout(self.layout)

    def enterEvent(self, *args, **kwargs):
        # self.frame.setFrameStyle(QFrame.Raised)
        pass

    # def leaveEvent(self, *args, **kwargs):
    # self.frame.setFrameStyle(QFrame.Plain)

    # def mousePressEvent(self, *args, **kwargs):
    # self.frame.setFrameStyle(QFrame.Sunken)

    def mouseDoubleClickEvent(self, *args, **kwargs):
        if os.path.isdir(self.path):
            self.parentwindow.libraryview.chdir(self.path)
        else:
            self.parentwindow.openfile(self.path)

class ScaledLabel(QtGui.QLabel):
    """
    A label that scales with the constrained dimension; for thumbnails
    """

    def __init__(self, image):
        super(ScaledLabel, self).__init__()
        self._pixmap = QtGui.QPixmap.fromImage(image)
        self._pixmap = QtGui.QPixmap(self._pixmap)

    def resizeEvent(self, event):
        self.setPixmap(self._pixmap.scaled(
            self.width(), self.height(),
            Qt.KeepAspectRatio))


# -*- coding: UTF-8 -*-

import os

import fabio
import numpy as np
import pyfits

# from nexpy.api import nexus as nx
from pyFAI import detectors
import pyFAI
import glob
import re
import writer
from xicam import debugtools, config
from pipeline.formats import TiffStack, CondensedTiffStack
from PySide import QtGui
from collections import OrderedDict
from pipeline import msg
# try:
#     import libtiff
# except IOError:
#     warnings.warn('libtiff not loaded; 3D tiffs cannot be read')

import numpy as nx

import detectors  # injects pyFAI with custom detectors

acceptableexts = ['.fits', '.edf', '.tif', '.tiff', '.nxs', '.hdf', '.cbf', '.img', '.raw', '.mar3450', '.gb', '.h5',
                  '.out', '.txt', '.npy']
imagecache = dict()


def loadsingle(path):
    return loadimage(path), loadparas(path)


def loadimage(path):
    data = None
    try:
        ext = os.path.splitext(path)[1]
        if ext in acceptableexts:
            # if ext in ['.nxs', '.hdf']:
            #     nxroot = nx.load(path)
            #     # print nxroot.tree
            #     if hasattr(nxroot, 'data'):
            #         if hasattr(nxroot.data, 'signal'):
            #             data = nxroot.data.signal
            #             return data
            #         else:
            #             return loadimage(str(nxroot.data.rawfile))
            # else:
                data = fabio.open(path).data
                return data
    except IOError:
        msg.logMessage('IO Error loading: ' + path,msg.ERROR)
    except TypeError:
        msg.logMessage('The selected file is not a type understood by fabIO.',msg.ERROR)

    return data


def readenergy(path):
    try:
        if os.path.splitext(path)[1] in acceptableexts:
            if os.path.splitext(path)[1] == '.fits':
                head = pyfits.open(path)
                # print head[0].header.keys()
                paras = scanparaslines(str(head[0].header).split('\r'))
                # print paras
            elif os.path.splitext(path)[1] in ['.nxs', '.hdf']:
                pass
                # nxroot = nx.load(path)
                # # print nxroot.tree
                # if hasattr(nxroot.data, 'signal'):
                # data = nxroot.data.signal
                #     return data, nxroot
                # else:
                #     print ('here:', nxroot.data.rawfile)
                #     return loadsingle(str(nxroot.data.rawfile))
                # print('here',data)

            else:
                pass
    except IOError:
        msg.logMessage('IO Error reading energy: ' + path,msg.ERROR)

    return None


def readvariation(path):

    try:
        nxroot = nx.load(path)
        return dict([[int(index), int(value)] for index, value in nxroot.data.variation])
    except IOError:
        msg.logMessage(('Could not load', path),msg.ERROR)
    except nx.NeXusError:
        msg.logMessage('No variation saved in file ' + path,msg.ERROR)

    return None


def merge_dicts(*dict_args):
    '''
    Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.
    '''
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result


def loadparas(path):
    try:
        extension = os.path.splitext(path)[1]
        if extension == '.fits':
            head = pyfits.open(path)
            # print head[0].header
            return head[0].header

        elif extension == '.edf':
            return fabio.open(path).header

        elif extension == '.gb':
            return {}

        elif extension in ['.nxs', '.hdf']:
            nxroot = nx.load(path)
            # print nxroot.tree
            return nxroot

        elif extension in ['.tif', '.img', '.tiff']:
            fimg = fabio.open(path)
            return fimg.header


    except IOError:
        msg.logMessage('Unexpected read error in loadparas',msg.ERROR)
    except IndexError:
        msg.logMessage('No txt file found in loadparas',msg.WARNING)
    return OrderedDict()

def loadstitched(filepath2, filepath1, data1=None, data2=None, paras1=None, paras2=None):
    if data1 is None or data2 is None or paras1 is None or paras2 is None:
        (data1, paras1) = loadsingle(filepath1)
        (data2, paras2) = loadsingle(filepath2)

    # DEFAULT TILING AT 733
    positionY1 = 0 if '_lo_' in filepath1 else 30 * .172
    positionY2 = 30 * .172 if '_hi_' in filepath2 else 0
    positionX1 = 0
    positionX2 = 0

    if config.activeExperiment.mapHeader('Detector Vertical') in paras1 and config.activeExperiment.mapHeader('Detector Vertical') in paras2 and \
                    config.activeExperiment.mapHeader('Detector Horizontal') in paras1 and config.activeExperiment.mapHeader('Detector Horizontal') in paras2:
        if float(paras1[config.activeExperiment.mapHeader('Detector Vertical')])!=0 and float(paras2[config.activeExperiment.mapHeader('Detector Vertical')])!=0:
            positionY1 = float(paras1[config.activeExperiment.mapHeader('Detector Vertical')])
            positionY2 = float(paras2[config.activeExperiment.mapHeader('Detector Vertical')])
            positionX1 = float(paras1[config.activeExperiment.mapHeader('Detector Horizontal')])
            positionX2 = float(paras2[config.activeExperiment.mapHeader('Detector Horizontal')])

    I1 = 1
    I2 = 1
    if config.activeExperiment.mapHeader('I1 AI') in paras1 and config.activeExperiment.mapHeader('I1 AI') in paras2:
        I1 = float(paras1[config.activeExperiment.mapHeader('I1 AI')])
        I2 = float(paras2[config.activeExperiment.mapHeader('I1 AI')])

    deltaX = round((positionX2 - positionX1) / 0.172)
    deltaY = round((positionY2 - positionY1) / 0.172)
    padtop2 = 0
    padbottom1 = 0
    padtop1 = 0
    padbottom2 = 0
    padleft2 = 0
    padright1 = 0
    padleft1 = 0
    padright2 = 0

    if deltaY < 0:
        padtop2 = int(abs(deltaY))
        padbottom1 = int(abs(deltaY))
    else:
        padtop1 = int(abs(deltaY))
        padbottom2 = int(abs(deltaY))

    if deltaX < 0:
        padleft2 = int(abs(deltaX))
        padright1 = int(abs(deltaX))

    else:
        padleft1 = int(abs(deltaX))
        padright2 = int(abs(deltaX))

    d2 = np.pad(data2, ((padtop2, padbottom2), (padleft2, padright2)), 'constant')
    d1 = np.pad(data1, ((padtop1, padbottom1), (padleft1, padright1)), 'constant')

    mask2 = np.pad(1 - finddetectorbyfilename(filepath2, data2).calc_mask(),
                      ((padtop2, padbottom2), (padleft2, padright2)),
                      'constant')
    mask1 = np.pad(1 - finddetectorbyfilename(filepath1, data1).calc_mask(),
                      ((padtop1, padbottom1), (padleft1, padright1)),
                      'constant')

    with np.errstate(divide='ignore',invalid='ignore'):
        data = (d1 / I1 + d2 / I2) / (mask2 + mask1) * (I1 + I2) / 2.
        data[np.isnan(data)] = 0
    return data, np.logical_or(mask2, mask1).astype(np.int)


def loadthumbnail(path):
    try:
        if os.path.splitext(path)[1] in ['.nxs', '.hdf']:
            nxroot = nx.load(pathtools.path2nexus(path))
            # print nxroot.tree
            if hasattr(nxroot, 'data'):
                if hasattr(nxroot.data, 'thumbnail'):
                    thumb = np.array(nxroot.data.thumbnail)
                    return thumb
    except IOError:
        msg.logMessage('IO Error loading: ' + path,msg.ERROR)
    except TypeError:
        msg.logMessage(('TypeError: path has type ', str(type(path))),msg.ERROR)

    return None


# @debugtools.timeit
# def finddetector(imgdata):
#     for name, detector in detectors.ALL_DETECTORS.iteritems():
#         if hasattr(detector, 'MAX_SHAPE'):
#             #print name, detector.MAX_SHAPE, imgdata.shape[::-1]
#             if detector.MAX_SHAPE == imgdata.shape[::-1]:  #
#                 detector = detector()
#                 mask = detector.calc_mask()
#                 print 'Detector found: ' + name
#                 return name, mask, detector
#         if hasattr(detector, 'BINNED_PIXEL_SIZE'):
#             #print detector.BINNED_PIXEL_SIZE.keys()
#             if imgdata.shape[::-1] in [tuple(np.array(detector.MAX_SHAPE) / b) for b in
#                                        detector.BINNED_PIXEL_SIZE.keys()]:
#                 detector = detector()
#                 mask = detector.calc_mask()
#                 print 'Detector found with binning: ' + name
#                 return name, mask, detector
#     raise ValueError('Detector could not be identified!')
#     return None, None, None

def finddetectorbyshape(shape):
    for name, detector in sorted(pyFAI.detectors.ALL_DETECTORS.iteritems()):
        # print detector
        # print 'det:',name, detector
        if hasattr(detector, 'MAX_SHAPE'):
            # print name, detector.MAX_SHAPE, imgdata.shape[::-1]
            if detector.MAX_SHAPE == shape[::-1]:  #
                detector = detector()
                msg.logMessage('Detector found: ' + name)
                return detector
        if hasattr(detector, 'BINNED_PIXEL_SIZE'):
            # print detector.BINNED_PIXEL_SIZE.keys()
            for binning in detector.BINNED_PIXEL_SIZE.keys():
                if shape[::-1] == tuple(np.array(detector.MAX_SHAPE) / binning):
                    detector = detector()
                    msg.logMessage('Detector found with binning: ' + name)
                    detector.set_binning(binning)
                    return detector
    return pyFAI.detectors.Detector(1, 1, max_shape=shape[::-1])
    raise ValueError('Detector could not be identified!')

def finddetectorbyfilename(path, data=None):
    if data is None:
        data = loadimage(path)

    dimg = diffimage(filepath=path, data=data)
    return dimg.detector


# def loadthumbnail(path):
# # nxpath = pathtools.path2nexus(path)
#
#     img=diffimage(filepath=path).thumbnail
#
#     return img


def loadpath(path):
    # Now returns data and mask
    if '_lo_' in path or '_hi_' in path:
        try:
            if '_lo_' in path:
                path2 = path.replace('_lo_', '_hi_')
            else:  # '_hi_' in path:
                path2 = path.replace('_hi_', '_lo_')
            return loadstitched(path, path2)
        except Exception as ex:
            msg.logMessage(('Stitching failed: ', ex.message),msg.ERROR)

    img = loadimage(path)

    # Do extra rotations/transposition
    if config.settings['Image Load Transpose']:
        img=img.transpose()
    if config.settings['Image Load Rotations']:
        img=np.rot90(img,config.settings['Image Load Rotations'])

    if not isinstance(img, tuple):
        mask = finddetectorbyshape(img.shape).calc_mask()
        if mask is None: mask = np.zeros_like(img)
        img = (img, 1-mask)
    return img


def loadxfs(path):
    return np.loadtxt(path, skiprows=16, converters={
        0: lambda s: int(s.split(':')[0]) * 60 * 60 + int(s.split(':')[1]) * 60 + int(s.split(':')[2])})


def convertto8bit(image):
    display_min = image.min()
    display_max = image.max()
    # image = np.array(image, copy=False)
    # image.clip(display_min, display_max, out=image)
    # image -= display_min
    np.true_divide(image, (display_max - display_min + 1) / 256., out=image, casting='unsafe')
    return image.astype(np.uint8)


def loadtiffstack(path):
    msg.logMessage(('Loading', path + '...'))
    data = np.swapaxes(libtiff.TIFF3D.open(path).read_image(), 0, 1)
    msg.logMessage('Sub-sampling array...')
    # data = convertto8bit(data)
    msg.logMessage(('Load complete. Size:', np.shape(data)))
    data = data[::4, ::4, ::4]
    return data.astype(np.float32)


def loadimageseries(pattern):
    msg.logMessage(('Loading', pattern + '...'))
    files = glob.glob(pattern)
    data = np.dstack([fabio.open(f).data for f in files])
    msg.logMessage('Log scaling data...')
    data = (np.log(data * (data > 0) + (data < 1)))
    msg.logMessage('Converting to 8-bit and re-scaling...')
    data = convertto8bit(data)
    msg.logMessage(('Load complete. Size:', np.shape(data)))
    return data


import integration, remesh, center_approx, variation, pathtools


class diffimage():
    def __init__(self, filepath=None, data=None, detector=None, experiment=None):
        """
        Image class for diffraction images that caches and validates cache
        :param filepath: str
        :param data: numpy.multiarray.ndarray
        :param detector: pyFAI.detectors.Detector
        :param experiment: xicam.config.experiment
        """

        msg.logMessage('diffimage is deprecated. Migrate this to diffimage2',msg.WARNING)

        msg.logMessage('Loading ' + unicode(filepath) + '...')

        self._data = data

        self.filepath = filepath
        self._detector = detector
        self._params = None
        self._thumb = None
        self._variation = dict()
        self._headers = None
        self._jpeg = None
        self.experiment = experiment
        if self.experiment is None:
            self.experiment = config.activeExperiment

        ### All object I want to save that depend on config parameters must be cached in here instead!!!!
        self.cache = dict()
        self.cachecheck = None

    def updateexperiment(self):
        # Force cache the detector
        _ = self.detector

        # Set experiment energy
        if 'Beamline Energy' in self.params:
            self.experiment.setvalue('Energy', self.params['Beamline Energy'])

    def checkcache(self):
        pass
        # compare experiment with cachecheck

    def invalidatecache(self):
        self.cache = dict()
        msg.logMessage('cache cleared')

    def cachedata(self):
        if self._data is None:
            if self.filepath is not None:
                try:
                    self._data, self.experiment.mask = loadpath(self.filepath)
                except IOError:
                    debugtools.frustration()
                    raise IOError('File moved, corrupted, or deleted. Load failed')



    @property
    def mask(self):
        if self.experiment.mask is not None:
            return self.experiment.mask
        else:
            return np.ones_like(self.data)

    @property
    def dataunrot(self):
        self.cachedata()
        return self._data

    @property
    def data(self):
        self.cachedata()
        data = self._data
        if self._data is not None:
            data = np.rot90(self._data, 3)
        return data

    @data.setter
    def data(self, data):
        self._data = data

    @property
    def params(self):
        if self._params is None:
            self._params = loadparas(self.paths[0])
        return self._params

    @property
    def detector(self):
        if self._detector is None:
            if self.data is not None:
                name, detector = self.finddetector()
            else:
                return None

            if detector is None:  # If none are found, use the last working one
                return self._detector

            self.detectorname = name
            mask = detector.calc_mask()
            self._detector = detector
            if detector is not None:
                if self.experiment is not None:
                    if mask is not None:
                        self.experiment.addtomask(np.rot90(1 - mask, 3))  # FABIO uses 0-valid mask
                    self.experiment.setvalue('Pixel Size X', detector.pixel1)
                    self.experiment.setvalue('Pixel Size Y', detector.pixel2)
                    self.experiment.setvalue('Detector', type(detector))
        return self._detector

    def finddetector(self):
        for name, detector in sorted(pyFAI.detectors.ALL_DETECTORS.iteritems()):
            # print detector
            # print 'det:',name, detector
            if hasattr(detector, 'MAX_SHAPE'):
                # print name, detector.MAX_SHAPE, imgdata.shape[::-1]
                if detector.MAX_SHAPE == self.data.shape[::-1]:  #
                    detector = detector()
                    msg.logMessage('Detector found: ' + name)
                    return name, detector
            if hasattr(detector, 'BINNED_PIXEL_SIZE'):
                # print detector.BINNED_PIXEL_SIZE.keys()
                for binning in detector.BINNED_PIXEL_SIZE.keys():
                    if self.data.shape[::-1] == tuple(np.array(detector.MAX_SHAPE) / binning):
                        detector = detector()
                        msg.logMessage('Detector found with binning: ' + name)
                        detector.set_binning(binning)
                        return name, detector
        return None, None
        raise ValueError('Detector could not be identified!')

    @detector.setter
    def detector(self, value):
        if type(value) == str:
            try:
                self._detector = pyFAI.detectors.ALL_DETECTORS[value]
            except KeyError:
                try:
                    self._detector = getattr(detectors, value)
                except AttributeError:
                    raise KeyError('Detector not found in pyFAI registry: ' + value)
        else:
            self._detector = value

    @property
    def params(self):
        if self._params is None:
            self._params = loadparas(self.filepath)
        return self._params

    @property
    def thumbnail(self):
        if self._thumb is None:
            self._thumb = loadthumbnail(self.filepath)
            if self._thumb is None:
                self._thumb = writer.thumbnail(self.data)
        return self._thumb

    @property
    def jpeg(self):
        if self._jpeg is None:
            self._jpeg = writer.jpeg(self.data.astype(np.uint8))
        return self._jpeg

    def iscached(self, key):
        return key in self.cache

    def cachedetector(self):
        _ = self.detector

    @property
    def cake(self):
        try:
            self.cachedetector()
            if not self.iscached('cake'):
                cake, x, y = integration.cake(self.data, self.experiment)
                cakemask, _, _ = integration.cake(np.ones_like(self.data), self.experiment)
                cakemask = cakemask > 0

                self.cache['cake'] = cake
                self.cache['cakemask'] = cakemask
                self.cache['cakeqx'] = x
                self.cache['cakeqy'] = y

            return self.cache['cake']
        except AttributeError as ex:
            msg.logMessage(ex.message,msg.ERROR)

    @property
    def remesh(self):
        if not self.iscached('remesh'):
            msg.logMessage(('headers:', self.headers),msg.DEBUG)
            # read incident angle
            if "Sample Alpha Stage" in self.headers:
                alphai = np.deg2rad(float(self.headers["Sample Alpha Stage"]))
            elif "Alpha" in self.headers:
                alphai = np.deg2rad(float(self.headers['Alpha']))
            elif "Sample Theta" in self.headers:
                alphai = np.deg2rad(float(self.headers['Sample Theta']))
            elif "samtilt" in self.headers:
                alphai = np.deg2rad(float(self.headers['samtilt']))
            else:
                alphai = self.queryAlphaI()
            if alphai is None:
                return self.data

            remeshdata, x, y = remesh.remesh(np.rot90(self.data, 1).copy(), self.filepath,
                                             self.experiment.getGeometry(), alphai)
            remeshmask, _, _ = remesh.remesh(np.rot90(self.mask).copy(), self.filepath,
                                             self.experiment.getGeometry(), alphai)

            self.cache['remesh'] = remeshdata
            self.cache['remeshmask'] = remeshmask > 0
            self.cache['remeshqx'] = x
            self.cache['remeshqy'] = y

        return self.cache['remesh']

    def queryAlphaI(self):
        alphai, ok = QtGui.QInputDialog.getDouble(None, u'Incident Angle', u'Enter incident angle (degrees):',
                                                  decimals=3)
        if alphai and ok:
            alphai = np.deg2rad(alphai)
            self.headers['Alpha'] = alphai
            return alphai

        return None

    def __del__(self):
        # TODO: do more here!
        # if self._data is not None:
        #    self.writenexus()
        pass

    @debugtools.timeit
    def writenexus(self):
        nxpath = pathtools.path2nexus(self.filepath)
        w = writer.nexusmerger(img=self._data, thumb=self.thumbnail, path=nxpath, rawpath=self.filepath,
                               variation=self._variation)
        w.run()

    def findcenter(self):
        # Auto find the beam center
        [x, y] = center_approx.center_approx(self.data)

        # Set the center in the experiment
        self.experiment.center = (x, y)

    def integrate(self, mode='', cut=None):
        ai = config.activeExperiment.getAI().getPyFAI()
        iscake = False
        return integration.radialintegratepyFAI(self.data, self.mask, ai, cut=cut)

    @debugtools.timeit  # 0.07s on Izanami
    def variation(self, operationindex, roi):
        if operationindex not in self._variation or roi is not None:
            nxpath = pathtools.path2nexus(self.filepath)
            if os.path.exists(nxpath) and roi is None:
                v = readvariation(nxpath)
                # print v
                if operationindex in v:
                    self._variation[operationindex] = v[operationindex]
                    msg.logMessage('successful variation load!')
                else:
                    prv = pathtools.similarframe(self.filepath, -1)
                    nxt = pathtools.similarframe(self.filepath, +1)
                    self._variation[operationindex] = variation.filevariation(operationindex, prv, self.dataunrot, nxt)
            else:
                prv = pathtools.similarframe(self.filepath, -1)
                nxt = pathtools.similarframe(self.filepath, +1)
                if roi is None:
                    self._variation[operationindex] = variation.filevariation(operationindex, prv, self.dataunrot, nxt)
                else:
                    v = variation.filevariation(operationindex, prv, self.dataunrot, nxt, roi)
                    return v
        return self._variation[operationindex]

    @property
    def headers(self):
        if self._headers is None:
            self._headers = loadparas(self.filepath)

        return self._headers

    @property
    def radialintegration(self):
        if 'radialintegration' in self.cache.keys():
            self.cache['radialintegration'] = integration.radialintegrate(self)

        return self.cache['radialintegration']

    def __getattr__(self, name):
        if name in self.cache:
            return self.cache[name]
        else:
            raise AttributeError('diffimage has no attribute: ' + name)


class imageseries():
    def __init__(self, paths, experiment=None):
        self.paths = dict()
        self.variation = dict()
        self.appendimages(paths)
        self.experiment = experiment
        self.roi = None
        self._thumbs = None
        self._dimgs = None
        self._jpegs = None

        if self.experiment is None:
            self.experiment = config.activeExperiment

    @property
    def dimgs(self):
        if self._dimgs is None:
            self._dimgs = [self.__getitem__(i) for i in range(self.__len__())]
        return self._dimgs

    def __len__(self):
        return len(self.paths)

    @property
    def xvals(self):
        return np.array(sorted(self.paths.keys()))

    def first(self):
        if len(self.paths) > 0:
            firstpath = sorted(list(self.paths.values()))[0]

            return diffimage(filepath=firstpath, experiment=self.experiment)
        else:
            return diffimage(data=np.zeros((2, 2)), experiment=self.experiment)

    def __getitem__(self, item):
        return self.getDiffImage(self.paths.keys()[item])

    def getDiffImage(self, key):
        # print self.paths.keys()

        return diffimage(filepath=self.paths[key], experiment=self.experiment)

    def appendimages(self, paths):
        for path in paths:
            frame = self.path2frame(path)
            if frame is None:
                continue

            self.variation[frame] = None
            self.paths[frame] = path

    def currentdiffimage(self):
        pass

    # @property
    # def roi(self):
    # return self._roi
    #
    # @roi.setter
    # def roi(self,value):
    # self._roi=value

    def scan(self, operationindex, roi=None):
        if len(self.paths) < 3:
            return None

        variation = dict()

        # get the first frame's profile
        keys = self.paths.keys()

        if roi is not None:
            roi = writer.thumbnail(roi.T)

        for key, index in zip(keys, range(self.__len__())):
            variationx = self.path2frame(self.paths[key])
            variationy = self.calcVariation(index, operationindex, roi)

            if variationy is not None:
                variation[variationx] = variationy

        return variation

    def calcVariation(self, i, operationindex, roi):
        if roi is None:
            roi = 1
        if i == 0:
            return None  # Prevent wrap-around with first variation

        thumbs = self.thumbs
        try:
            if operationindex == 7:
                return variation.variationoperators.operations.values()[operationindex](self[i].data, i, roi)
            return variation.variationoperators.operations.values()[operationindex](thumbs, i, roi)
        except IndexError as ex:
            msg.logMessage(('Skipping index:', i),msg.WARNING)
        return None

    @property
    @debugtools.timeit
    def thumbs(self):
        if self._thumbs is None:
            self._thumbs = [dimg.thumbnail for dimg in self.dimgs]
        return self._thumbs

    @property
    def jpegs(self):
        if self._jpegs is None:
            self._jpegs = jpegimageset([dimg.jpeg for dimg in self.dimgs])

        return self._jpegs

    @staticmethod
    def path2frame(path):
        try:
            expr = '(?<=_)[\d]+(?=[_.])'
            return int(re.findall(expr, os.path.basename(path))[-1])

        except ValueError:
            msg.logMessage(('Path has no frame number:', path),msg.ERROR)

        return None


from PIL import Image


class jpegimageset():
    def __init__(self, jpegs):
        self.jpegs = jpegs

    @property
    def dtype(self):
        return np.uint8

    @property
    def max(self):
        return np.max(Image.open(self.jpegs[0]))

    @property
    def min(self):
        return np.min(Image.open(self.jpegs[0]))

    @property
    def ndim(self):
        return 3

    @property
    def shape(self):
        return (len(self.jpegs), np.shape(Image.open(self.jpegs[0]))[0], np.shape(Image.open(self.jpegs[0]))[1])

    @property
    def size(self):
        return len(self.jpegs) * np.product(np.size(Image.open(self.jpegs[0])))

    def __getitem__(self, item):
        msg.logMessage(('item:', item),msg.DEBUG)
        if type(item) in (int, np.int64):
            return self.jpegs[item]
        else:
            return np.array([self.jpegs[i] for i in item])


class StackImage(object):
    """
    Class for displaying a Image Stack in a pyqtgraph ImageView and be able to scroll through the various Images
    """

    ndim = 3

    def __init__(self, filepath=None, data=None):
        super(StackImage, self).__init__()
        self._rawdata = None
        self.filepath = filepath

        if filepath is not None:
            if (isinstance(filepath, list) and len(filepath) == 1):
                filepath = filepath[0]
            else:
                self.fabimage = fabio.open(filepath)
        elif data is not None:
            self.fabimage = data
        else:
            if filepath is None and data is None:
                raise ValueError('Either data or path to file must be provided')
        # throw error if loading with fabio fails
        if not self.fabimage:
            raise IOError("Unable to detect file format for this dataset.")

        self.header = self.fabimage.header

        self._framecache = dict()
        self._cachesize = 2
        self.currentframe = 0

        raw = self.rawdata
        self.dtype = raw.dtype
        self.max = np.max(raw)
        self.min = np.min(raw)
        self.shape = len(self.fabimage), raw.shape[0], raw.shape[1]
        self.size = np.product(self.shape)

    @property
    def rawdata(self):
        # 'Permanently' cached
        if self._rawdata is None:
            self._rawdata = self._getframe()
        return self._rawdata

    def transpose(self, ax):
        # TODO: find a good way to do this
        # TODO: annoying because of the way hdfs are stored
        return self

    def asVolume(self, level=1):
        for i, j in enumerate(range(0, self.shape[0], level)):
            img = self._getimage(j)[::level, ::level]
            if i == 0:  # allocate array:
                shape = (np.ceil(float(self.shape[0]) / level), img.shape[0], img.shape[1])
                vol = np.empty(shape, dtype=self.rawdata.dtype)
            vol[i] = img
        return vol

    def _getframe(self, frame=None):  # keeps 3 frames in cache at most
        if frame is None: frame = self.currentframe
        if type(frame) is list and type(frame[0]) is slice:
            frame = 0  # frame[1].step
        self.currentframe = frame
        # print self._framecache
        if frame not in self._framecache:
            # del the first cached item
            if len(self._framecache) > self._cachesize: del self._framecache[self._framecache.keys()[0]]
            self._framecache[frame] = self._getimage(frame)
        return self._framecache[frame]

    def _getimage(self, frame):
        return self.fabimage.getframe(frame)

    def invalidatecache(self):
        self.cache = dict()

    # This needs more thought to get some slices out of there
    def __getitem__(self, item):
        return self._getframe(item)

    def __del__(self):
        try:
            self.fabimage.close()
        except ValueError:
            pass


class diffimage2(object):
    def __init__(self, detector=None, experiment=None):

        """
        Image class for diffraction images that caches and validates cache
        :param detector: pyFAI.detectors.Detector
        :param experiment: xicam.config.experiment
        """

        msg.logMessage('Loading...')

        self.logscale = True
        self.remeshmode = False
        self.cakemode = False
        self.radialsymmetrymode = False
        self.mirrorsymmetrymode = False

        self._rawdata = None
        self._detector = detector
        self._params = None
        self._thumb = None
        self._variation = dict()
        self._headers = None
        self._jpeg = None
        self.experiment = experiment
        if self.experiment is None:
            self.experiment = config.activeExperiment

        config.activeExperiment.setvalue('Incidence Angle (GIXS)', np.rad2deg(self.getAlphaI()))

        ### All object I want to save that depend on config parameters must be cached in here instead!!!!
        self.cache = dict()
        self.cachecheck = None

    def __len__(self):
        if hasattr(self, 'filepaths'):
            return len(self.filepaths)
        else:
            return 1

    def updateexperiment(self):
        # Force cache the detector
        _ = self.detector

        # Set experiment energy
        if 'Beamline Energy' in self.params:
            self.experiment.setvalue('Energy', self.params['Beamline Energy'])
        elif 'mono' in self.params:
            self.experiment.setvalue('Energy', self.params['mono'])

    def checkcache(self):
        pass
        # compare experiment with cachecheck

    def invalidatecache(self):
        self.cache = dict()
        msg.logMessage('cache cleared')

    def cachedata(self):
        if self._rawdata is None:
            if self.filepath is not None:
                try:
                    self._rawdata, self.experiment.mask = loadpath(self.filepath)
                except IOError:
                    debugtools.frustration()
                    raise IOError('File moved, corrupted, or deleted. Load failed')

    @property
    def mask(self):
        if self.experiment.mask is not None:
            return self.experiment.mask
        else:
            return np.ones_like(self.rawdata)


    @property
    def transformdata(self):
        # Not cached
        img = self.rawdata
        if self.radialsymmetrymode:
            img = self.radialsymmetryfill(img)
        elif self.mirrorsymmetrymode:
            img = self.mirrorsymmetryfill(img)

        if self.cakemode:
            img = self.cake(img, self.mask)
        elif self.remeshmode:
            img = self.remesh(img, self.mask)

        return img

    @property
    def transformmask(self):
        img = self.mask
        if self.radialsymmetrymode:
            img = self.radialsymmetryfill(img)
        elif self.mirrorsymmetrymode:
            img = self.mirrorsymmetryfill(img)

        if self.cakemode:
            img = self.cakemask
        elif self.remeshmode:
            img = self.remesh(img, self.mask) > 0

        return img

    def radialsymmetryfill(self, img):
        centerx = config.activeExperiment.center[0]
        centery = config.activeExperiment.center[1]
        symimg = np.rot90(img.copy(), 2)

        xshift = -(img.shape[0] - 2 * centerx)
        yshift = -(img.shape[1] - 2 * centery)
        symimg = np.roll(symimg, int(xshift), axis=0)
        symimg = np.roll(symimg, int(yshift), axis=1)

        marginmask = config.activeExperiment.mask

        x, y = np.indices(img.shape)
        padmask = ((yshift < y) & (y < (yshift + img.shape[1])) & (xshift < x) & (x < (xshift + img.shape[0])))

        img = img * marginmask + symimg * padmask * (1 - marginmask)
        return img

    def mirrorsymmetryfill(self, img):
        centerx = config.activeExperiment.getvalue('Center X')
        symimg = np.flipud(img.copy())
        self.imtest(symimg)
        xshift = -(img.shape[1] - 2 * centerx)
        symimg = np.roll(symimg, int(xshift), axis=0)
        self.imtest(symimg)
        marginmask = config.activeExperiment.mask
        self.imtest(marginmask)

        x, y = np.indices(img.shape)
        padmask = ((xshift < x) & (x < (xshift + img.shape[1])))
        self.imtest(padmask)
        self.imtest(symimg * padmask * (1 - marginmask))
        img = img * marginmask + symimg * padmask * (1 - marginmask)
        return img

    @property
    def params(self):
        if self._params is None:
            self._params = loadparas(self.paths[0])
        return self._params

    @property
    def detector(self):
        if self._detector is None:
            if self.rawdata is not None:
                name, detector = self.finddetector()
            else:
                return None

            if detector is None:  # If none are found, use the last working one
                return self._detector

            self.detectorname = name
            mask = detector.calc_mask()
            self._detector = detector
            if detector is not None:
                if self.experiment is not None:
                    if mask is not None:
                        self.experiment.addtomask(np.rot90(1 - mask, 3))  # FABIO uses 0-valid mask
                    self.experiment.setvalue('Pixel Size X', detector.pixel1)
                    self.experiment.setvalue('Pixel Size Y', detector.pixel2)
                    self.experiment.setvalue('Detector', type(detector))
        return self._detector

    def finddetector(self):
        for name, detector in sorted(pyFAI.detectors.ALL_DETECTORS.iteritems()):
            if hasattr(detector, 'MAX_SHAPE'):
                # print name, detector.MAX_SHAPE, self.rawdata.shape[::-1]
                if detector.MAX_SHAPE == self.rawdata.shape[::-1]:  #
                    detector = detector()
                    msg.logMessage('Detector found: ' + name)
                    return name, detector
            if hasattr(detector, 'BINNED_PIXEL_SIZE'):
                # print detector.BINNED_PIXEL_SIZE.keys()
                for binning in detector.BINNED_PIXEL_SIZE.keys():
                    if self.rawdata.shape[::-1] == tuple(np.array(detector.MAX_SHAPE) / binning):
                        detector = detector()
                        msg.logMessage('Detector found with binning: ' + name)
                        detector.set_binning(binning)
                        return name, detector
        return None, None

    @detector.setter
    def detector(self, value):
        if type(value) == str:
            try:
                self._detector = pyFAI.detectors.ALL_DETECTORS[value]
            except KeyError:
                try:
                    self._detector = getattr(detectors, value)
                except AttributeError:
                    raise KeyError('Detector not found in pyFAI registry: ' + value)
        else:
            self._detector = value

    @property
    def params(self):
        if self._params is None:
            self._params = loadparas(self.filepath)
        return self._params

    @property
    def thumbnail(self):
        if self._thumb is None:
            self._thumb = loadthumbnail(self.filepath)
            if self._thumb is None:
                self._thumb = writer.thumbnail(self.data)
        return self._thumb

    @property
    def jpeg(self):
        if self._jpeg is None:
            self._jpeg = writer.jpeg(self.data.astype(np.uint8))
        return self._jpeg

    def iscached(self, key):
        return key in self.cache

    def cachedetector(self):
        _ = self.detector

    def cake(self, img, mask):
        self.cachedetector()
        if not self.iscached('cake'):
            cake, x, y = integration.cake(img, self.experiment, mask=mask)
            cakemask, _, _ = integration.cake(np.ones_like(img), self.experiment, mask=mask)
            cakemask = cakemask > 0

            self.cache['cake'] = cake
            self.cache['cakemask'] = cakemask
            self.cache['cakeqx'] = x
            self.cache['cakeqy'] = y

        return self.cache['cake']

    def getAlphaI(self):
        alphai = 0
        if "Sample Alpha Stage" in self.headers:
            alphai = np.deg2rad(float(self.headers["Sample Alpha Stage"]))
        elif "Alpha" in self.headers:
            alphai = np.deg2rad(float(self.headers['Alpha']))
        elif "Sample Theta" in self.headers:
            alphai = np.deg2rad(float(self.headers['Sample Theta']))
        elif "samtilt" in self.headers:
            alphai = np.deg2rad(float(self.headers['samtilt']))
        else:
            msg.logMessage('No incidence angle found in headers. Consider mapping key to internal variable.')
            alphai=0
        return alphai

    def remesh(self, img, mask):
        if not self.iscached('remesh'):
            # read incident angle

            alphai = np.deg2rad(config.activeExperiment.getvalue('Incidence Angle (GIXS)'))
            msg.logMessage('Using incidence angle value: ' + str(alphai))

            remeshdata, x, y = remesh.remesh(np.rot90(img).copy(), self.filepath,
                                             self.experiment.getAI(), alphai)
            remeshmask, _, _ = remesh.remesh(np.rot90(mask).copy(), self.filepath,
                                             self.experiment.getAI(), alphai)

            self.cache['remesh'] = remeshdata
            self.cache['remeshmask'] = remeshmask > 0
            self.cache['remeshqx'] = x
            self.cache['remeshqy'] = y

        return self.cache['remesh']

    def findcenter(self):
        # Auto find the beam center
        [x, y] = center_approx.center_approx(self.rawdata)

        # Set the center in the experiment
        self.experiment.center = (x, y)

    @property
    def headers(self):
        if self._headers is None:
            self._headers = loadparas(self.filepath)

        return self._headers

    def queryAlphaI(self):
        alphai, ok = QtGui.QInputDialog.getDouble(None, u'Incident Angle', u'Enter incident angle (degrees):',
                                                  decimals=3)
        if alphai and ok:
            alphai = np.deg2rad(alphai)
            self.headers['Alpha'] = alphai
            return alphai

        return None

    @property
    def displaydata(self):
        # Not cached
        if self.logscale:
            return np.log(self.transformdata * (self.transformdata > 0) + (self.transformdata < 1))
        return self.transformdata

    def asarray(self):
        return self.displaydata

    def view(self, t):
        if t is np.ndarray:
            return self.displaydata

    def transpose(self,ax):
        if ax != [0,1,2]:
            msg.logMessage('Conflict with newstyle pyqtgraph imageview; see line 669',level=msg.ERROR)
        return self

    def __getitem__(self, item):
        return self.displaydata[item]

    def __getattr__(self, name):
        if name in self.cache:
            return self.cache[name]
        else:
            raise AttributeError('diffimage has no attribute: ' + name)


# class singlefilediffimage2(diffimage2):
#     ndim=2
#     def __init__(self, filepath, detector=None, experiment=None):
#         self.filepath = filepath
#         super(singlefilediffimage2, self).__init__(detector=detector, experiment=experiment)
#         return self.cache['remesh']
#
#     def queryAlphaI(self):
#         alphai, ok = QtGui.QInputDialog.getDouble(None, u'Incident Angle', u'Enter incident angle (degrees):',
#                                                   decimals=3)
#         if alphai and ok:
#             alphai = np.deg2rad(alphai)
#             self.headers['Alpha'] = alphai
#             return alphai
#
#         return None
#
#
#     def __del__(self):
#         # TODO: do more here!
#         # if self._data is not None:
#         #    self.writenexus()
#         pass
#
#     @debugtools.timeit
#     def writenexus(self):
#         nxpath = pathtools.path2nexus(self.filepath)
#         w = writer.nexusmerger(img=self._rawdata, thumb=self.thumbnail, path=nxpath, rawpath=self.filepath,
#                                variation=self._variation)
#         w.run()
#
#     def findcenter(self):
#         # Auto find the beam center
#         [x, y] = center_approx.center_approx(self.rawdata)
#
#         # Set the center in the experiment
#         self.experiment.center = (x, y)
#
#     def integrate(self, mode='', cut=None):
#         ai = config.activeExperiment.getAI().getPyFAI()
#         iscake = False
#         return integration.radialintegratepyFAI(self.data, self.mask, ai, cut=cut)
#
#     @debugtools.timeit  #0.07s on Izanami
#     def variation(self, operationindex, roi):
#         if operationindex not in self._variation or roi is not None:
#             nxpath = pathtools.path2nexus(self.filepath)
#             if os.path.exists(nxpath) and roi is None:
#                 v = readvariation(nxpath)
#                 #print v
#                 if operationindex in v:
#                     self._variation[operationindex] = v[operationindex]
#                     print 'successful variation load!'
#                 else:
#                     prv = pathtools.similarframe(self.filepath, -1)
#                     nxt = pathtools.similarframe(self.filepath, +1)
#                     self._variation[operationindex] = variation.filevariation(operationindex, prv, self.dataunrot, nxt)
#             else:
#                 prv = pathtools.similarframe(self.filepath, -1)
#                 nxt = pathtools.similarframe(self.filepath, +1)
#                 if roi is None:
#                     print prv, self.dataunrot, nxt
#                     self._variation[operationindex] = variation.filevariation(operationindex, prv, self.dataunrot, nxt)
#                 else:
#                     v = variation.filevariation(operationindex, prv, self.dataunrot, nxt, roi)
#                     return v
#         return self._variation[operationindex]
#
#     @property
#     def headers(self):
#         if self._headers is None:
#             self._headers = loadparas(self.filepath)
#
#         return self._headers
#
#     @property
#     def radialintegration(self):
#         if 'radialintegration' in self.cache.keys():
#             self.cache['radialintegration'] = integration.radialintegrate(self)
#
#         return self.cache['radialintegration']
#
#     def view(self,t):
#         if t is np.ndarray:
#             return self.displaydata
#

#
#     def __getattr__(self, name):
#        if name in self.cache:
#            return self.cache[name]
#        else:
#            raise AttributeError('diffimage has no attribute: ' + name)


# each diffimage class should implement:
# rawdata, transformdata, displaydata

class singlefilediffimage2(diffimage2):
    ndim = 2

    def __init__(self, filepath, detector=None, experiment=None):
        self.filepath = filepath
        super(singlefilediffimage2, self).__init__(detector=detector, experiment=experiment)

    @property
    def rawdata(self):
        # 'Permanently' cached
        if self._rawdata is None:
            rawdata, mask = loadpath(self.filepath)
            self._rawdata = np.rot90(rawdata, 3)
            if mask is not None and (self.experiment.mask is None or self.experiment.mask.shape!=mask.shape): self.experiment.mask = np.rot90(mask, 3)
        return self._rawdata

    def implements(self, t):
        if t == 'MetaArray': return True


class multifilediffimage2(diffimage2):
    ndim = 3

    def __init__(self, filepaths, detector=None, experiment=None):
        self.filepaths = sorted(list(filepaths))
        self._currentframe = 0
        super(multifilediffimage2, self).__init__(detector=detector, experiment=experiment)
        self._framecache = dict()
        self._xvals = None

        self.dtype = self.rawdata.dtype
        self.max = np.max(self.rawdata)
        self.min = np.min(self.rawdata)
        self.shape = len(filepaths), self.rawdata.shape[0], self.rawdata.shape[1]
        self.size = np.product(self.shape)

    @property
    def currentframe(self):
        if self._currentframe is None: return 0
        return self._currentframe

    @currentframe.setter
    def currentframe(self, frame):
        if frame != self._currentframe: self._rawdata = None
        self._currentframe = frame


    @property
    def headers(self):
        if self._headers is None:
            self._headers = loadparas(self.filepaths[self.currentframe])

        return self._headers

    def iHeaders(self, i):
        return loadparas(self.filepaths[i])

    def xvals(self, _):
        if self._xvals is None:
            timekey = config.activeExperiment.headermap['Timeline Axis']
            if timekey:
                self._xvals = np.array([float(self.iHeaders(i)[timekey]) for i in range(len(self.filepaths))])
            else:
                self._xvals = np.arange(len(self.filepaths))
        return self._xvals

    def first(self):
        if len(self.filepaths) > 0:
            firstpath = self.filepaths[0]
            return diffimage(filepath=firstpath, experiment=self.experiment)
        else:
            return diffimage(data=np.zeros((2, 2)), experiment=self.experiment)

    @property
    def rawdata(self):
        # 'Permanently' cached
        if self._rawdata is None:
            self._rawdata = np.rot90(loadimage(self.filepaths[self.currentframe]), 3)
        return self._rawdata

    @property
    def transformdata(self):
        # 'Temporary' cached
        if self.cakemode:
            if 'cake' not in self.cache:
                self.cake()
            return self.cache['cake']

        elif self.remeshmode:
            if 'remesh' not in self.cache:
                self.remesh()
            return self.cache['remesh']

        return self.rawdata

    @property
    def displaydata(self):
        # Not cached
        # msg.logMessage(('applyinglog:', self.logscale),msg.DEBUG)
        if self.logscale:
            return np.log(self.transformdata * (self.transformdata > 0) + (self.transformdata < 1))
        return self.transformdata

    def _getframe(self, frame=None):  # keeps 3 frames in cache at most
        # print 'frame:',frame
        if frame is None: frame = self.currentframe
        if type(frame) is list: frame = frame[2].step
        self.currentframe = frame
        if not frame in self._framecache:
            if len(self._framecache) > 3: del self._framecache[
                sorted(self._framecache.keys())[0]]  # del the first cached item
            self._framecache[frame] = self.displaydata
        return self._framecache[frame]

    def calcVariation(self, i, operationindex, roi):
        if roi is None:
            roi = 1
        if i == 0:
            return None  # Prevent wrap-around with first variation

        try:
            var=variation.variationoperators.operations.values()[operationindex](self, i, roi)
            t=self.xvals('')[i]
            return t, var
        except IndexError as ex:
            msg.logMessage(('Skipping index:', i),msg.WARNING)
        return None

    def __getitem__(self, item):
        return self._getframe(item)


class datadiffimage2(singlefilediffimage2):
    ndim = 2

    def __init__(self, data, detector=None, experiment=None):
        super(datadiffimage2, self).__init__(filepath=None, detector=detector, experiment=experiment)
        self._rawdata = np.rot90(data, 3)

        # False scale rawdata to avoid log issues
        if self._rawdata.max() <= 1:
            self._rawdata *= 2 ** 32

    @property
    def headers(self):
        return dict()


class stackdiffimage2(diffimage2):
    ndim = 3

    def __init__(self, filepath, detector=None, experiment=None):
        self.filepath = filepath
        super(stackdiffimage2, self).__init__(detector=detector, experiment=experiment)

        self.fabimage = fabio.open(filepath)

        self._framecache = dict()
        self.currentframe = 0

        raw = self.rawdata
        self.dtype = raw.dtype
        self.max = np.max(raw)
        self.min = np.min(raw)
        self.shape = len(self.fabimage), raw.shape[0], raw.shape[1]
        self.size = np.product(self.shape)

    @property
    def rawdata(self):
        # 'Permanently' cached
        if self._rawdata is None:
            self._rawdata = self._getframe()
        return self._rawdata

    @property
    def transformdata(self):
        # 'Temporary' cached
        if self.cakemode:
            if 'cake' not in self.cache:
                self.cake()
            return self.cache['cake']

        elif self.remeshmode:
            if 'remesh ' not in self.cache:
                self.remesh()
            return self.cache['remesh']

        return self.rawdata

    @property
    def displaydata(self):
        # Not cached
        if self.logscale:
            return np.log(self.transformdata * (self.transformdata > 0) + (self.transformdata < 1))
        return self.transformdata

    def _getframe(self, frame=None):  # keeps 3 frames in cache at most
        if frame is None: frame = self.currentframe
        if type(frame) is list and type(frame[0]) is slice:
            frame = frame[1].step
        frame = min(frame, len(self))
        self.currentframe = frame
        if frame not in self._framecache:
            if len(self._framecache) > 2: del self._framecache[self._framecache.keys()[0]]  # del the first cached item
            self._framecache[frame] = np.rot90(self.fabimage.getframe(frame).data, 3)
        return self._framecache[frame]

    def __getitem__(self, item):
        return self._getframe(item)


def loaddiffimage(src):
    if type(src) in [unicode, str]:
        return singlefilediffimage2(src)
    elif type(src) is list and len(src) == 1 and os.path.splitext(src[0])[-1] == '.h5':
        return stackdiffimage2(src[0])
    elif type(src) is list and len(src) == 1:
        return singlefilediffimage2(src[0])
    elif type(src) is list:
        return multifilediffimage2(src)
    elif src is None:
        return None



#! /usr/bin/env python

import glob
import os
import time

import numpy as np
import fabio
from scipy import optimize
from scipy import signal

import saxs_calibration
import peakfindingrem


def approx_width(r):
    """
    linearly varies the peak width between GISAXS, where peaks are
    thinner to GIWAXS, where peaks are fewer but wider.
    """
    return (0.047 * r + 1.8261)


def tophat2(radius, scale=1):
    """
    convolution kernel, revolved to form a ring, with Mexican Hat
    profile along the radial direction.

    radius : peak position along the radius
    scale  : magnification factor
    """
    width = approx_width(radius)
    N = np.round(radius) + 3 * round(width) + 1
    x = np.arange(-N, N)
    x, y = np.meshgrid(x, x)
    t = np.sqrt(x ** 2 + y ** 2) - radius
    s = width
    a = scale / np.sqrt(2 * np.pi) / s ** 3
    w = a * (1 - (t / s) ** 2) * np.exp(-t ** 2 / s ** 2 / 2.)
    return w


def calc_R(x, y, xc, yc):
    """ calculate the distance of each 2D points from the center (xc, yc) """
    return np.sqrt((x - xc) ** 2 + (y - yc) ** 2)


def f(c, x, y):
    """ calculate the algebraic distance between the data points and the mean circle centered at c=(xc, yc) """
    Ri = calc_R(x, y, *c)
    return Ri - Ri.mean()


def fitpointstocircle(cnt):
    """Fit a Nx2 array of points to a circle; return center, radius, and fit residue"""

    # In case the shape is strange (i.e. extra dimensions) reshape it
    cnt = np.array(cnt.reshape((-1, 2)))

    # if the contour doesn't have enough points, give up
    if cnt.shape[0] < 3:
        return None, None, None, None

    # separate the points into x and y arrays
    x = cnt[:, 0]
    y = cnt[:, 1]

    # fit algorithm for circle from cookbook
    x_m = np.mean(x)
    y_m = np.mean(y)
    center_estimate = x_m, y_m
    center, ier = optimize.leastsq(f, center_estimate, args=(x, y))
    xc, yc = center
    Ri = calc_R(x, y, *center)
    R = Ri.mean()
    residu = np.sum((Ri - R) ** 2)
    return xc, yc, R, residu


# @debug.timeit
# def oldcenter_approx(img):
#     """
#     Find the center of scattering image; maskfit parameters are the Canny threshold parameters for sharp edges and
#     the Hough lines transform threshold; edgefitp parameters are the Canny thresholds for finding edges after masking
#     """
#     # newcenter_approx(img, experiment)
#     #return newcenter_approx(img, log=False)
#
#     try:
#         # Rescale brightness of the image with log depth
#         with np.errstate(divide='ignore', invalid='ignore'):
#             img = np.log(img) / np.log(np.max(img)) * 255
#
#         # Convert to 8bit image for the HoughCircles procedure
#         img = np.uint8(img)
#
#         # Histogram levels and equalize
#         img = cv2.equalizeHist(img)
#
#         # Draw for demo
#         if demo:
#             cv2.imshow('step?', cv2.resize(img, (0, 0), fx=.5, fy=.5))
#             cv2.waitKey(0)
#             cv2.destroyAllWindows()
#
#         # Add a 4 pixel gaussian blur to denoise the line search image
#         linesimg = cv2.GaussianBlur(img.copy(), (0, 0), 2)
#         linesimg = cv2.bilateralFilter(linesimg, -1, 10, 5)
#
#         # Do Canny edge search with mask parameters (looking for sharp edges to mask)
#         linesimg = cv2.Canny(linesimg, maskfitp[0], maskfitp[1])
#
#         kernel = np.ones((15, 15), np.uint8)
#         sharpmask = 1 - (cv2.dilate(cv2.morphologyEx(linesimg.copy(), cv2.MORPH_CLOSE, kernel), kernel) / 255)
#
#         # print sharpmask
#
#         # Draw for demo
#         if demo:
#             cv2.imshow('step?', cv2.resize(linesimg, (0, 0), fx=.5, fy=.5))
#             cv2.waitKey(0)
#             cv2.destroyAllWindows()
#
#             cv2.imshow('step?', cv2.resize(255 * sharpmask, (0, 0), fx=.5, fy=.5))
#             cv2.waitKey(0)
#             cv2.destroyAllWindows()
#
#         # Do Hough lines transform looking for lines at least as long as 1/4 the image width
#         #lines = cv2.HoughLines(linesimg, 1, np.pi / 180, 100, img.shape[0] / 4, maskfitp[2])
#
#         # Make an array for the mask
#         #linesmask = np.ones_like(img)
#
#         # for rho, theta in lines[0]:  # For each line found
#         #     a = np.cos(theta)
#         #     b = np.sin(theta)
#         #     x0 = a * rho
#         #     y0 = b * rho
#         #     x1 = int(x0 + 10000 * (-b))
#         #     y1 = int(y0 + 10000 * (a))
#         #     x2 = int(x0 - 10000 * (-b))
#         #     y2 = int(y0 - 10000 * (a))
#         #
#         #     # Mask out the line with a 20px bar
#         #     cv2.line(linesmask, (x1, y1), (x2, y2), (0, 0, 0), 20)
#         #
#         # # Draw for demo
#         # if demo == True:
#         #     cv2.imshow('step?', cv2.resize(linesmask * 255, (0, 0), fx=.5, fy=.5))
#         #     cv2.waitKey(0)
#         #     cv2.destroyAllWindows()
#
#         # Add a 3px gaussian blur to denoise the scattering image
#         img = cv2.GaussianBlur(img, (0, 0), 3)
#
#         # Add bilateral filtering to further denoise, preserving edges
#         img = cv2.bilateralFilter(img, -1, 10, 10)
#
#         # Cut the image at half its z-depth; this creates very good edges for Canny
#         img = img * 3
#
#         # Do Canny edge search to find rings; also apply the mask now
#         img = cv2.Canny(img, edgefitp[0], edgefitp[1]) * sharpmask
#
#         if demo:
#             cv2.imshow('detected circles', cv2.resize(img, (0, 0), fx=.5, fy=.5))
#             # wait to kill
#             cv2.waitKey(0)
#             cv2.destroyAllWindows()
#
#         # Make array to save circles
#         circles = []
#
#         # Find contours created by Canny; Each contour is a list of all chained points
#         contours, hierarchy = cv2.findContours(img.copy(), cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
#
#         for (h, cnt), cntindx in zip(enumerate(contours), range(
#                 contours.__len__())):  # For each contour; Zipped with a index to track each circle
#             # Fit the contour to a circle
#             x, y, R, residu = fitpointstocircle(cnt)
#
#             if x is not None and img.shape[0] > R > 200:  # If a circle was found and it has reasonable radius
#                 # Make a mask to select all additional points on the circle; If the circle was fit with only an arc,
#                 # this will include points further out
#                 mask = np.zeros_like(img)
#                 cv2.circle(mask, (int(x), int(y)), int(R), 255, 2)
#
#                 # Calculate how much of the full circle was found; this is the score for that circle
#                 coverage = (np.sum(img * mask) / (np.pi * 2 * R))
#
#                 # Add the circle to the list of circles found
#                 circles.append([x, y, R, coverage, cntindx])
#
#         # Make the circles array a nummpy array now that we're done building it
#         circles = np.array(circles)
#
#         # Sort and take the best circle (most coverage)
#         bestcircle = circles[circles[:, 3].argsort()[-1]]
#
#         # Draw the best guess for demo
#         if demo:
#             cv2.circle(img, (int(bestcircle[0]), int(bestcircle[1])), int(bestcircle[2]), 100, 5)
#
#         # Get all similar contours, looking for any circles that are within 20px radius and center
#         Rrange = 15
#         posrange = 15
#         similarcontours = []
#         for circle in circles:  # For each circle
#             # If the circle is close enough
#             if (bestcircle[2] - Rrange < circle[2] < bestcircle[2] + Rrange) \
#                     and (bestcircle[1] - posrange < circle[1] < bestcircle[1] + posrange) \
#                     and (bestcircle[0] - posrange < circle[0] < bestcircle[0] + posrange):
#                 # Add its points to a new binary image
#                 similarcontours.append(contours[int(circle[4])][:, 0, :])
#
#                 # Draw similar circles for demo
#                 if demo:
#                     cv2.circle(img, (int(circle[0]), int(circle[1])), int(circle[2]), 100, 5)
#
#         # Compress the array with vstack
#         similarcontours = np.vstack(similarcontours)
#
#         # Find the best circle fit from the new contour set
#         bestcircle = fitpointstocircle(similarcontours)
#
#         # Draw the finalized circle for demo
#         if demo:
#             cv2.circle(img, (int(bestcircle[0]), int(bestcircle[1])), int(bestcircle[2]), 255, 5)
#             cv2.imshow('detected circles', cv2.resize(img, (0, 0), fx=.5, fy=.5))
#             # wait to kill
#             cv2.waitKey(0)
#             cv2.destroyAllWindows()
#
#         # Return only the center
#         return bestcircle[0:2]
#     except Exception as ex:
#         template = "An exception of type {0} occured in center_approx. Arguments:\n{1!r}"
#         message = template.format(type(ex).__name__, ex.args)
#         print message
#         print "The center was unable to be found for an image."
#         return None

from xicam import debugtools


@debugtools.timeit
def center_approx(img, mask, log=False):

    if log:
        # Rescale brightness of the image with log depth
        img = img.astype(np.float)
        with np.errstate(divide='ignore', invalid='ignore'):
            img = np.log(img * (img > 0) + 1)

    con = signal.fftconvolve(img.astype(int)*mask*(img>0), img.astype(int)*mask*(img>0)) / np.sqrt(signal.fftconvolve(np.ones_like(img), np.ones_like(img)))

    cen = np.array(np.unravel_index(con.argmax(), con.shape)) / 2.
    return cen


# def tth_ellipse(geometry, d_spacing):
# '''
# Returns a matplotlib.patches.Ellipse to plot the given d-spacing ring.
# '''
#     tth = 2 * np.arcsin(geometry.get_wavelength() / (2e-10 * d_spacing))
#
#     geo_dict = geometry.getFit2D()
#     sdd = geo_dict['directDist'] / (0.001 * geo_dict['pixelX'])
#     tilt = np.deg2rad(geo_dict['tilt'])
#     rotation = np.deg2rad(geo_dict['tiltPlanRotation'])
#
#     c_plus = (sdd * np.sin(tth)) / np.sin(np.pi / 2 - tilt - tth)
#     c_minus = (sdd * np.sin(tth)) / np.sin(np.pi / 2 + tilt - tth)
#     elli_h = (sdd * np.sin(tth)) / np.sin(np.pi / 2 - tth)
#     elli_w = (c_plus + c_minus) / 2.0
#
#     x_pos = (geo_dict['centerX'] - c_minus + elli_w) - geo_dict['centerX']
#     elli_x = geo_dict['centerX'] + x_pos * np.cos(rotation)
#     elli_y = geo_dict['centerY'] + x_pos * np.sin(rotation)
#
#     return matplotlib.patches.Ellipse((elli_x, elli_y),
#                                       elli_w * 2,
#                                       elli_h * 2,
#                                       np.rad2deg(rotation))


######################################################################################
######  REMI
###
#


def gisaxs_center_approx(img, log=False):
    img = img.astype(np.float)
    if log:
        # Rescale brightness of the image with log depth
        with np.errstate(divide='ignore', invalid='ignore'):
            img = np.log(img + 3) - np.log(3)

    x = 0
    xcenter = 0
    y = 10000
    ycenter = 0
    for i in range(0, img.shape[1]):
        if x <= sum(img[:, i]):
            x = sum(img[:, i])
            xcenter = i
        else:
            pass

    q = 4 * sum(img[img.shape[0] - 5, :])
    i = 0
    x = np.sum(img[:, :150], axis=1)
    for i in range(1, np.size(x)):
        if x[i] == 0:
            x[i] = x[i - 1]
        else:
            pass
    t = np.size(x) - 20
    x = signal.convolve(signal.convolve(x[:t], signal.gaussian(7, std=4)), [1, -1])
    # plt.plot(x)
    # plt.show()

    i = 0
    while (y != np.min(x)):
        y = x[i]
        ycenter = i - 6  # 6 because correct spread form convolution gaussian and derivation
        i = i + 1

    cen = (xcenter, ycenter)
    return cen



















    #
    ###
    ######  REMI


#########################################################################################################

def refinecenter(dimg):
    imgcopy = dimg.rawdata.T
    # Refine calibration
    # d-spacing for Silver Behenate
    d_spacings = np.array([58.367, 29.1835, 19.45567, 14.59175, 11.6734, 9.72783, 8.33814, 7.29587, 6.48522, 5.8367])

    geometry = dimg.experiment.getGeometry()

    # print 'Start parameter:'
    # print geometry.getFit2D()

    fit_param = ['center_x', 'center_y', 'distance', 'rotation', 'tilt']
    fit_thread = saxs_calibration.FitThread(geometry, d_spacings, imgcopy, fit_param, 40)
    fit_thread.start()
    while fit_thread.is_alive():
        #print fit_thread.status
        time.sleep(.1)

    # print 'Final parameter:'
    #print geometry.getFit2D()
    #print fit_thread.get_info()

    #pylab.imshow(np.log(img), interpolation='none')

    # for d_spacing in d_spacings:
    #    ellipse = tth_ellipse(geometry, d_spacing)
    #    ellipse.set_fc('none')
    #    ellipse.set_ec('red')
    #    pylab.gca().add_patch(ellipse)

    # for circle in fit_thread.get_circle_patches():
    #    pylab.gca().add_patch(circle)

    #pylab.show()

    return geometry.getFit2D()['centerX'], geometry.getFit2D()['centerY']

# def testimg(img, scale=0.5):
#     # Draw for demo
#     demo = False
#     if demo:
#         cv2.imshow('step?', cv2.resize((img * 255.0 / img.max()).astype(np.uint8), (0, 0), fx=.5, fy=.5))
#         cv2.waitKey(0)
#         cv2.destroyAllWindows()


# if __name__ == "__main__":
# i = 0
# for imgpath in glob.glob(os.path.join("../../saxswaxs/AgBs", '*.edf')):
#         i += 1
#         if i < 0:
#             continue
#
#         print "Opening", imgpath
#
#         # read image
#         img = fabio.open(imgpath).data
#
#         outputimg = img.copy()
#
#         # print new center approximation; Add demo=True to see it work!
#         circle = center_approx(img, demo=True)
#         if circle is not None:
#             outputimg = np.uint8(outputimg)
#             outputimg = cv2.cvtColor(outputimg, cv2.COLOR_GRAY2BGR)
#             cv2.circle(outputimg, (int(circle[0]), int(circle[1])), int(circle[2]), (255, 0, 0), 3)
#             cv2.circle(outputimg, (int(circle[0]), int(circle[1])), 10, (255, 0, 0), 10)
#
#             cv2.imwrite(imgpath + "_center.png", cv2.resize(outputimg, (0, 0), fx=.3, fy=.3))
